-- ===========================================================================
--	City Banner Manager
-- ===========================================================================

print("Loading CityBannerManager.lua from Loyalty++ version "..GlobalParameters.Lpp_VERSION_MAJOR.."."..GlobalParameters.Lpp_VERSION_MINOR.."."..GlobalParameters.Lpp_VERSION_PATCH);

include( "CivilizationIcon" );
include( "InstanceManager" );
include( "SupportFunctions" );
include( "Civ6Common" );
include( "LuaClass" );
include( "CitySupport" );                     -- Get CityData


-----------------------------------------------------------------------------------------
--    FILE:     SaveLoad.lua
--  Gedemon (2017)
--  Loading/Saving simple tables with the game file using Pickle Table for serialization
--  http://lua-users.org/wiki/PickleTable
-----------------------------------------------------------------------------------------

----------------------------------------------
-- Pickle.lua
-- A table serialization utility for lua
-- Steve Dekorte, http://www.dekorte.com, Apr 2000
-- Freeware
----------------------------------------------

function pickle(t)
  return Pickle:clone():pickle_(t)
end

Pickle = {
  clone = function (t) local nt={}; for i, v in pairs(t) do nt[i]=v end return nt end
}

function Pickle:pickle_(root)
  if type(root) ~= "table" then
    error("can only pickle tables, not ".. type(root).."s")
  end
  self._tableToRef = {}
  self._refToTable = {}
  local savecount = 0
  self:ref_(root)
  local s = ""

  while #(self._refToTable) > savecount do
    savecount = savecount + 1
    local t = self._refToTable[savecount]
    s = s.."{\n"
    for i, v in pairs(t) do
        s = string.format("%s[%s]=%s,\n", s, self:value_(i), self:value_(v))
    end
    s = s.."},\n"
  end

  return string.format("{%s}", s)
end

function Pickle:value_(v)
  local vtype = type(v)
  if     vtype == "string" then return string.format("%q", v)
  elseif vtype == "number" then return v
  elseif vtype == "boolean" then return tostring(v)
  elseif vtype == "table" then return "{"..self:ref_(v).."}"
  else --error("pickle a "..type(v).." is not supported")
  end
end

function Pickle:ref_(t)
  local ref = self._tableToRef[t]
  if not ref then
    if t == self then error("can't pickle the pickle class") end
    table.insert(self._refToTable, t)
    ref = #(self._refToTable)
    self._tableToRef[t] = ref
  end
  return ref
end

----------------------------------------------
-- unpickle
----------------------------------------------

function unpickle(s)
  if type(s) ~= "string" then
    error("can't unpickle a "..type(s)..", only strings")
  end
  local gentables = loadstring("return "..s)
  local tables = gentables()

  for tnum = 1, #(tables) do
    local t = tables[tnum]
    local tcopy = {}; for i, v in pairs(t) do tcopy[i] = v end
    for i, v in pairs(tcopy) do
      local ni, nv
      if type(i) == "table" then ni = tables[i[1]] else ni = i end
      if type(v) == "table" then nv = tables[v[1]] else nv = v end
      t[i] = nil
      t[ni] = nv
    end
  end
  return tables[1]
end



-----------------------------------------------------------------------------------------
-- Load / Save
-- Using Civ6 GameConfiguration
-----------------------------------------------------------------------------------------

-- save
function SaveTableToSlot(t, sSlotName)
    local s = pickle(t)
    GameConfiguration.SetValue(sSlotName, s)
end

-- load
function LoadTableFromSlot(sSlotName)
    local s = GameConfiguration.GetValue(sSlotName)
    if s then
        local t = unpickle(s)
        return t
    else
        print("ERROR: can't load table from ".. tostring(sSlotName))
    end
end

function Initialize()
    ExposedMembers.SaveTableToSlot = SaveTableToSlot
    ExposedMembers.LoadTableFromSlot = LoadTableFromSlot
end
Initialize()


-- ===========================================================================
--  Initialize
-- ===========================================================================



function InitializeWarExhaustion(value,playerid)              -- initialize all player War Exhaustion level
    Warexhaustion = LoadTableFromSlot("Warexhaustion")  
    if Warexhaustion == nil then
    Warexhaustion = {}
        print("warexhaustion got initialized")
    end
    if playerid == "initialize" then
        for id, players in pairs(Players) do
            local Warexhaustiontest = Warexhaustion[id]
                if Warexhaustiontest == nil then
                    Warexhaustion[id] = value
                end
        end
    else
        Warexhaustion[playerid] = value
    end
end

InitializeWarExhaustion(0,"initialize")



function InitializeCityReligionMod(owner, ID)              -- initialize all player cities ReligionMod
    
    
    if not ReligionMod then
    ReligionMod = {}
    tooltipReligionMod = {}
    print("ReligionMod got initialized")
    end
     
    if owner == nil or ID == nil then
        
    else
    local kOwnerCityId = owner.."."..ID
    ReligionMod[kOwnerCityId] = 0
    tooltipReligionMod[kOwnerCityId] = ""
    end
end

Events.CityInitialized.Add(				InitializeCityReligionMod );
InitializeCityReligionMod()



function InitializeRevoltType()                               -- Initialize city last revolt turn
    
    
    CityLastRevolt = LoadTableFromSlot("CityLastRevolt")  
    if CityLastRevolt == nil then
    CityLastRevolt = {}  
    end
    print("CityLastRevolt got initialized")
end

InitializeRevoltType()




function InitializeGovernmentMod ()                          -- Initioalize GovernmentMod
    
    
    GovernmentMod = LoadTableFromSlot("GovernmentMod")
    if GovernmentMod == nil then
        GovernmentMod = {}
        print("GovernmentMod got initialized")
    end
    if GovernmentReligionMultiplier == nil then
        GovernmentReligionMultiplier = {}
    end
    for id, players in pairs(Players) do
        local GovernmentModnil = GovernmentMod[id]
        if GovernmentModnil == nil then
            GovernmentMod[id] = 0
            GovernmentReligionMultiplier[id] = 1
        end
    end
end

InitializeGovernmentMod()

local CurrentTurn = -1

function Currentturn ()                          -- turn count
 CurrentTurn = Game.GetCurrentGameTurn()
return CurrentTurn;
end
GameEvents.OnGameTurnStarted.Add(Currentturn)



-- ===========================================================================
--  War Exhaustion
-- ===========================================================================


function SetWarexhaustionScripts(playerid,value)                    -- send War exhaustion value to Script and add warexhaustion to playerUI
        
    
        Warexhaustion[playerid] = Warexhaustion[playerid] + value
    if Warexhaustion[playerid] <= -300 then -- war exhaustion level can't be less than -300 
        Warexhaustion[playerid] = -300
    elseif Warexhaustion[playerid] >= 0 then -- war exhaustion level can't be less more than 0
        Warexhaustion[playerid] = 0     
    end
    LuaEvents.ChangePlayerWarexhaustion(playerid, Warexhaustion[playerid]);       -- create lua event to link with my script (thx Gedemon)
    SaveTableToSlot(Warexhaustion,"Warexhaustion")
end






    
    
function round56(num, numDecimalPlaces) -- round function
  local mult = 10^(numDecimalPlaces or 0)
  return math.floor(num * mult + 0.5) / mult
end




function OnDiplomacyDeclareWar2( eAttackingPlayerid, eDefendingPlayerid, eWarType ) -- Gain War Exhaustion when declaring war depending on warmonger
        
        local attackingplayer =  Players[eAttackingPlayerid] 
        local iWarmongerPoints = attackingplayer:GetDiplomacy():ComputeDOWWarmongerPoints(eDefendingPlayerid, eWarType);
        local szWarmongerLevel = attackingplayer:GetDiplomacy():GetWarmongerLevel(-iWarmongerPoints);
        if szWarmongerLevel == "LOC_WARMONGER_LEVEL_NONE" then                -- if no warmong
        SetWarexhaustionScripts(eAttackingPlayerid,-5)
        elseif szWarmongerLevel == "LOC_WARMONGER_LEVEL_LIGHT" then           -- if light warmong
        SetWarexhaustionScripts(eAttackingPlayerid,-20)
        elseif szWarmongerLevel == "LOC_WARMONGER_LEVEL_MODERATE" then        -- if moderate warmong
        SetWarexhaustionScripts(eAttackingPlayerid,-30)
        elseif szWarmongerLevel == "LOC_WARMONGER_LEVEL_HEAVY" then           -- if heavy warmong  
        SetWarexhaustionScripts(eAttackingPlayerid,-50)
        elseif szWarmongerLevel == "LOC_WARMONGER_LEVEL_SEVERE" then          -- if severe warmong
        SetWarexhaustionScripts(eAttackingPlayerid,-75)
        elseif szWarmongerLevel == "LOC_WARMONGER_LEVEL_EGREGIOUS" then       -- if egregious warmong
        SetWarexhaustionScripts(eAttackingPlayerid,-100)
        end
end

LuaEvents.APlayerDeclaredWar.Add(	OnDiplomacyDeclareWar2 );   -- lua event  in declarewar popup.lua

function IsAtWar(pPlayer)                                   -- return true if major player is at war with another major player else return false
    for id, players in pairs(Players) do
        if id ~= pPlayer then      
            local otherPlayer =  Players[id]  -- all players exept player
            local pPlayer1 = Players[pPlayer]  -- player
            if (not pPlayer1:IsBarbarian()) and pPlayer1:WasEverAlive() and (not otherPlayer:IsBarbarian()) and otherPlayer:WasEverAlive() then  
	               local pPlayerDiplomaticAI = pPlayer1:GetDiplomaticAI();
	               local iState  = pPlayerDiplomaticAI:GetDiplomaticStateIndex(id);

	                   -- Do not assume ordering of the database, so get the hash and compare the known values.
	               local kStateEntry = GameInfo.DiplomaticStates[iState];
	               local eState = kStateEntry.Hash;

	               if  eState == DiplomaticStates.WAR and otherPlayer:IsAlive() then 
		              return true;
                    end
            end
        end
	end
return false;
end

function IsAtWarWithWho(pPlayer)                                   -- return with who the player is at war with
    for id, players in pairs(Players) do
        if id ~= pPlayer then      
            local otherPlayer =  Players[id]  -- all players exept player
            local pPlayer1 = Players[pPlayer]  -- player
            if (not pPlayer1:IsBarbarian()) and pPlayer1:WasEverAlive() and (not otherPlayer:IsBarbarian()) and otherPlayer:WasEverAlive() then  
	               local pPlayerDiplomaticAI = pPlayer1:GetDiplomaticAI();
	               local iState  = pPlayerDiplomaticAI:GetDiplomaticStateIndex(id);

	                   -- Do not assume ordering of the database, so get the hash and compare the known values.
	               local kStateEntry = GameInfo.DiplomaticStates[iState];
	               local eState = kStateEntry.Hash;

	               if  eState == DiplomaticStates.WAR and otherPlayer:IsAlive() then 
		              return id;
                    end
            end
        end
	end
end

function AddWarExhaustionlevel (PlayerID)                              -- War exhaustion per turn from war or peace and occupied and assieged city
            if PlayerID == nil then
                print("ERROR 335 PlayerID IS NIL")
            else
                local pPlayer1 = Players[PlayerID]
                local pWinnerConfig2 = PlayerConfigurations[PlayerID];
                local civiliname2 = pWinnerConfig2:GetCivilizationDescription()
                local Warexhaustion2 = Warexhaustion[PlayerID]
                    if Warexhaustion2 == nil then
                        Warexhaustion[PlayerID] = 0
                    end
                if  IsAtWar(PlayerID) and Warexhaustion[PlayerID] >= -300 then  -- war exhaustion level can't be less than -300 
                    SetWarexhaustionScripts(PlayerID,-1)  -- war exhaustion per turn from war
                    local EnemyID = IsAtWarWithWho(PlayerID)
                    local pEnemy = Players[EnemyID]
                    for _, pCity in pEnemy:GetCities():Members() do
                        if pCity ~= nil and pCity:IsOccupied() and pCity:GetOwnerBeforeOccupation() == PlayerID then  -- war exhaustion per turn per city occupied by the enemy
                            SetWarexhaustionScripts(PlayerID,-1)
                        end
                    end
                    for _, pCity in pPlayer1:GetCities():Members() do
                        if pCity ~= nil then
                            local cityDistricts = pCity:GetDistricts()
                            for i,district in cityDistricts:Members() do
                                local districtInfo:table = GameInfo.Districts[district:GetType()];
                                local districtName = districtInfo.Name
                                print(districtName)
                                if districtName == "LOC_DISTRICT_CITY_CENTER_NAME" and district:IsUnderSiege() then -- war exhaustion per turn per city under siege
                                    print("yay")
                                    SetWarexhaustionScripts(PlayerID,-1)
                                end
                            end
                        end
                    end
                elseif  (not IsAtWar(PlayerID)) and Warexhaustion[PlayerID] < 0 then -- war exhaustion level can't be less than 0
                    SetWarexhaustionScripts(PlayerID,3) -- war exhaustion per turn from peace
                end
                 LuaEvents.ChangeLoyalty(PlayerID);
            end
           
end  
        
GameEvents.PlayerTurnStarted.Add( AddWarExhaustionlevel )


function IsAtWarWith2(pPlayer1ID,pPlayer2ID)                  -- return true if both player are at war with each other else return false
            local pPlayer1 =  Players[pPlayer1ID]
            local pPlayer2 = Players[pPlayer2ID] 
            if (not pPlayer1:IsBarbarian()) and pPlayer1:WasEverAlive() and (not pPlayer2:IsBarbarian()) and pPlayer2:WasEverAlive() and pPlayer1:IsMajor() and pPlayer2:IsMajor() then  
                   local pPlayerDiplomaticAI = pPlayer1:GetDiplomaticAI();
	               local iState  = pPlayerDiplomaticAI:GetDiplomaticStateIndex(pPlayer2ID);

	                   -- Do not assume ordering of the database, so get the hash and compare the known values.
	               local kStateEntry = GameInfo.DiplomaticStates[iState];
	               local eState = kStateEntry.Hash;

	               if  eState == DiplomaticStates.WAR then 
		              return true;
                   end
            end
return false;
end

function OnCombat( combatResult )                           -- add war exhaustion after combat 
    local attacker = combatResult[CombatResultParameters.ATTACKER]
    local defender = combatResult[CombatResultParameters.DEFENDER]
    local attackerid  = attacker[CombatResultParameters.ID].player -- get attacker player ID
    local defenderid  = defender[CombatResultParameters.ID].player -- get defender player ID
    local pPlayer1 = Players[attackerid] 
    local pPlayer2 = Players[defenderid]
    local AttackerDamageDone  = defender[CombatResultParameters.DAMAGE_TO] -- damge done by attacker
    local DefenderDamageDone  = attacker[CombatResultParameters.DAMAGE_TO] -- damge done by Defender
    local AttackerWarexhaustionLoss = round56 (DefenderDamageDone/15,0)
    local DefenderWarexhaustionLoss = round56 (AttackerDamageDone/15,0) -- damagetaken/10
    if IsAtWarWith2(attackerid,defenderid) and Warexhaustion[attackerid] >= -300 then
        SetWarexhaustionScripts(attackerid, -AttackerWarexhaustionLoss )  -- war exhaustion per turn from combat
        if (attacker[CombatResultParameters.ID].type == ComponentType.UNIT) and (attacker[CombatResultParameters.FINAL_DAMAGE_TO] >= attacker[CombatResultParameters.MAX_HIT_POINTS]) then
            SetWarexhaustionScripts(attackerid, -3 )  -- war exhaustion per turn from unit killed for attacker
        end
    end
    if IsAtWarWith2(attackerid,defenderid) and Warexhaustion[defenderid] >= -300 then
        SetWarexhaustionScripts(defenderid, -DefenderWarexhaustionLoss )  -- war exhaustion per turn from combat)
        if (defender[CombatResultParameters.ID].type == ComponentType.UNIT) and (defender[CombatResultParameters.FINAL_DAMAGE_TO] >= defender[CombatResultParameters.MAX_HIT_POINTS]) then
        SetWarexhaustionScripts(defenderid, -3 )  -- war exhaustion per turn from unit killed for defender
        end
    end

end

Events.Combat.Add( OnCombat )


function exhaustionOnUnitCaptured (currentUnitOwner, unit, owningPlayer, capturingPlayer)  -- add war exhaustion if your unit get captured
    OldOwner = Players[currentUnitOwner] 
    SetWarexhaustionScripts(currentUnitOwner, -3 ) -- war exhaustion per turn from unit captured for currentowner

end


Events.UnitCaptured.Add(exhaustionOnUnitCaptured);

-- ===========================================================================
--  RELIGION MOD
-- ===========================================================================


function SetReligionModScripts(value,playerid,CityID,playerturn)                    -- send ReligionMod value to Script and add ReligionModUI    
        local kOwnerCityId = playerid.."."..CityID  
        local  ReligionModtest = ReligionMod[kOwnerCityId]
        if type(value) == "number" then
            ReligionMod[kOwnerCityId] =  value
            local ReligionModScripts =   ReligionMod[kOwnerCityId]
            local pPlayer = Players[playerid]
            local pCity = pPlayer:GetCities():FindID(CityID)
            local CityName = pCity:GetName()
            if playerturn ~=nil and playerturn == playerid then
            LuaEvents.ChangePlayerReligionMod(value,playerid,CityName);    -- create lua event to link with my script (thx Gedemon)
            end
        else
            print("ERROR VALUE IS"..type(value))
        end
        
 
end



function SetReligionModPerHeretics(playerturn)                           -- Is surely x3 too long but work :D
   local currentEra:number = Game.GetEras():GetCurrentEra();  
   local currentEraName:string = GameInfo.Eras[currentEra].Name;
   
   local eramultiplier = 1
    
    if currentEraName == "LOC_ERA_ANCIENT_NAME" then                     -- set era multiplier
        eramultiplier = 1
        
    elseif currentEraName == "LOC_ERA_CLASSICAL_NAME" then
        eramultiplier = 1
        
    elseif currentEraName == "LOC_ERA_MEDIEVAL_NAME" then
        eramultiplier = 1.5
        
    elseif currentEraName == "LOC_ERA_RENAISSANCE_NAME" then
        eramultiplier = 1.5
        
    elseif currentEraName == "LOC_ERA_INDUSTRIAL_NAME" then
        eramultiplier = 1
        
    elseif currentEraName == "LOC_ERA_MODERN_NAME" then
        eramultiplier = 0.5
        
    elseif currentEraName == "LOC_ERA_ATOMIC_NAME" then
        eramultiplier = 0.5
        
    elseif currentEraName == "LOC_ERA_INFORMATION_NAME" then
        eramultiplier = 0.1
        
    elseif type(currentEraName) == "string" then
        print("Error unknow era "..currentEraName)
    else
        print("Error  currentEraName is "..type(currentEraName))
        
    end
    

    
    
    
    for playerid, pPlayer in pairs(Players) do
        if pPlayer ~= nil then 
        local playerCities = pPlayer:GetCities(); 
        local GovernmentMultiplier = GovernmentReligionMultiplier[playerid]
        if GovernmentMultiplier == nil then
                GovernmentMultiplier = 1
        end
        local TotalReligionMultiplier = GovernmentMultiplier * eramultiplier
        
            if playerCities ~= nil then
            
                for _, pCity in playerCities:Members() do
                    if pCity ~= nil then
                        local Value = 0
                        local tooltip = ""
                        local pCityName = pCity:GetName()
                        local CityID = pCity:GetID()
                        local pCityReligion = pCity:GetReligion()
                        local pPlayerReligion = pPlayer:GetReligion()
                    
                        if pCityReligion ~= nil then
                            local religionsInCity = pCityReligion:GetReligionsInCity();  
                            local MajorReligion = pCity:GetReligion():GetMajorityReligion()
                            local MajorPlayerReligion = pPlayerReligion:GetReligionInMajorityOfCities()
                            local ownPlayerReligion = -5
                            local MajorPlayerReligionexeptpagan = -6   
                            if pPlayerReligion ~= nil then
                                    local OwnPlayerReligionType	= pPlayerReligion:GetReligionTypeCreated();
                                    
                                    if OwnPlayerReligionType ~= -1 then
                                        ownPlayerReligion = OwnPlayerReligionType
                                 
                                    end
                                    
                                    local MajorPlayerReligionName = Game.GetReligion():GetName(MajorPlayerReligion)
                                    tooltip = tooltip.."Era multiplier is "..eramultiplier.." Government multiplier is "..GovernmentMultiplier.."[NEWLINE]Your current Multiplier is"..TotalReligionMultiplier.."[NEWLINE]" 
                                    if MajorPlayerReligion ~= -1 then
                                        MajorPlayerReligionexeptpagan = MajorPlayerReligion
                                        tooltip = tooltip.."[NEWLINE] Your Empire Follow a religion called "..MajorPlayerReligionName.." City who do not follow this religion are considered Heretical"
                                    else
                                        tooltip = tooltip.."[NEWLINE] Your Empire is Pagan: All religions are considered heretical "
                                    end
                            end
                        
                            local MajorReligionExeptPagan = -7
                            local MajorReligionName = ""
                            if MajorReligion ~= nil then
                                 MajorReligionName = Game.GetReligion():GetName(MajorReligion)
                                if MajorReligionName ~= "" then
                                    MajorReligionExeptPagan = MajorReligion
                                end
                            end
                            if MajorReligion ~= MajorPlayerReligion and MajorReligionExeptPagan ~= ownPlayerReligion   then
                                        Value = Value - 50 * TotalReligionMultiplier
                                        if MajorReligionName ~= "" then
                                            tooltip = tooltip.."[NEWLINE] ".. -50 * TotalReligionMultiplier .." This city Follows a religion called "..MajorReligionName
                                        else
                                            tooltip = tooltip.."[NEWLINE] ".. -50 * TotalReligionMultiplier .." This city is pagan "
                                        end
                            end
                            if MajorReligionExeptPagan == MajorPlayerReligionexeptpagan or MajorReligionExeptPagan == ownPlayerReligion  then
                                        Value = Value + 50 * TotalReligionMultiplier
                                        tooltip = tooltip.."[NEWLINE] +".. 50 * TotalReligionMultiplier .." This city is not considered Heretical cause it Follows a religion called "..MajorReligionName
                            end 
                                
                                
                             local OwnPlayerReligionName  = Game.GetReligion():GetName(OwnPlayerReligion)
                                
                            if MajorReligionName ~= "" and  ownPlayerReligion == MajoReligion then
                                    tooltip = tooltip.."[NEWLINE][NEWLINE] Your city follow your religion called "..MajorReligionName..", all other religions and pagans are considered heretical:"
                            elseif MajorReligionName ~= "" then
                                    tooltip = tooltip.."[NEWLINE][NEWLINE] Your city follow a religion called "..MajorReligionName.." and you created a religion called "..OwnPlayerReligionName..", all other religions and pagans are considered heretical: "
                            else
                                    tooltip = tooltip.."[NEWLINE][NEWLINE] Your city is pagan, all religion are considered heretical: "    
                            end   
                            for CityReligionid, CityReligion in pairs(religionsInCity) do
                                    
                                    local Followers = CityReligion.Followers
                                    local pressure = CityReligion.Pressure
                                    local ReligionID = CityReligion.Religion
                                    local ReligionIDName = Game.GetReligion():GetName(ReligionID)
                                     
                                    
                                    if ReligionID ~= MajorReligion and ReligionID ~= ownPlayerReligion and Followers ~= 0  then -- Loyalty lost per turn for every Heretics
                                        Value = Value + Followers * -10 * TotalReligionMultiplier
                                        if ReligionIDName ~= "" then
                                            tooltip = tooltip.."[NEWLINE] "..Followers * -10 * TotalReligionMultiplier .." There are "..Followers.." Heretics who follow a religion called "..ReligionIDName
                                        else
                                            tooltip = tooltip.."[NEWLINE] "..Followers * -10 * TotalReligionMultiplier .." There are "..Followers.." pagans "
                                        end
                                    end
                                    if (ReligionID == MajorReligion or ReligionID == ownPlayerReligion) and ReligionIDName ~= "" and Followers ~= 0 then -- Loyalty Gain per turn for every Follower of good religion
                                        Value = Value + Followers * 10 * TotalReligionMultiplier
                                        if ReligionID == ownPlayerReligion then
                                            tooltip = tooltip.."[NEWLINE] +"..Followers * 10 * TotalReligionMultiplier .." There are "..Followers.." Followers of your religion "..ReligionIDName
                                        else
                                            tooltip = tooltip.."[NEWLINE] +"..Followers * 10 * TotalReligionMultiplier .." There are "..Followers.." Followers of Dominant religion in City "..ReligionIDName
                                        end    
                                    end
                            end
                            local kOwnerCityId = playerid.."."..CityID 
                            SetReligionModScripts(Value,playerid,CityID,playerturn)
                            tooltipReligionMod[kOwnerCityId] = tooltip
                        end
                    end
                 end
            end
        end                      
    end        
end

GameEvents.PlayerTurnStarted.Add( SetReligionModPerHeretics )


-- ===========================================================================
--  Revolt Chooser
-- ===========================================================================

function GetFreeWorkedPlot(pCityPlots, pCity)
    local Allplots = {}
    local count = 0
    for _, plotID in ipairs(pCityPlots) do
        local plots = Map.GetPlotByIndex(plotID)
        local x = plots:GetX()
        local y = plots:GetY()
        local plot = pCity:GetCitizens():IsPlotWorked(x,y)
        if plot and not plots:IsUnit() and not plots:IsCity() then 
        Allplots[count] = plots
        count = count+1
        end
    end
    return Allplots
end


function RevoltType(  player   )                                                                -- chance of revolt + chose revolt type function
    local playerwarexhaustion =  Warexhaustion[player]                                          -- player WEP
    local pPlayer = Players[player]
    local ReligionMOdValue = 0
    local CityLastRevolt2 = -40
    local PlayerGovernmentMod = GovernmentMod[player]
    
    
    for _, pCity in pPlayer:GetCities():Members() do                                             -- for each cities of player
        if pCity ~= nil then
            local CityName = pCity:GetName()                                                     -- CityName
            local CityId = pCity:GetID()
            local kOwnerCityId = player.."."..pCity:GetID()
            local pCityData = GetCityData( pCity )
            ReligionMOdValue = ReligionMod[kOwnerCityId]                                   -- City Religion loyalty
            if type(ReligionMOdValue) ~= "number" then
                    ReligionMOdValue = 0
                    ReligionMod[kOwnerCityId] = 0
            end
       
            
            
            if type(playerwarexhaustion) ~= "number" then                                   
                    playerwarexhaustion = 0
                    Warexhaustion[player] = 0
            end
            
            
            
                CityLastRevolt2 = CityLastRevolt[kOwnerCityId]                                    -- City last revolt turn
             if type(CityLastRevolt2) ~= "number" then
                CityLastRevolt2 = -40
                CityLastRevolt[kOwnerCityId] = -40           
            end
            
            if type(PlayerGovernmentMod) ~= "number" then
                PlayerGovernmentMod = 0
                CGovernmentMod[player] = 0           
            end
            
            
            if CityLastRevolt2 + 30 < Currentturn() then
              
            
            
            
            
                
                local pCityCulturalIdentity = pCity:GetCulturalIdentity();      
                local LoyaltyPerTurn = pCityCulturalIdentity:GetLoyaltyPerTurn() + playerwarexhaustion + ReligionMOdValue -- City Loyalty loss pt
                local Cityloyalty = pCityCulturalIdentity:GetLoyalty() + LoyaltyPerTurn;                   -- City CurrentLoyalty
                if Cityloyalty < 1000 or LoyaltyPerTurn <= 0 then
                    local ChanceOfRevolt = -LoyaltyPerTurn / Cityloyalty * 1000                         -- Chance of revolt
                    local dice = math.random(0,1000)                                                      -- Dice
                    if dice <= ChanceOfRevolt then
                    
                        CityLastRevolt[kOwnerCityId] = Currentturn()                                        -- set last revolt for city
                        SaveTableToSlot(CityLastRevolt,"CityLastRevolt")
                        local pCityPlots = Map.GetCityPlots():GetPurchasedPlots( pCity);
                        local CityPopulation = pCity:GetPopulation();                                      -- Citypop
                        LuaEvents.restoreLoyalty( player, CityName, Cityloyalty );                                     -- restoreLoyalty
                    
                    
                        for Barbarianid,pBarbarian in pairs(Players) do                                        
                            if pBarbarian:IsBarbarian() then 
                                    local AllPlots = GetFreeWorkedPlot(pCityPlots, pCity)                        
                                for _,WorkedPlot in pairs(AllPlots) do                                    -- For each worked plot 1 barbare spawn
                
                                    if WorkedPlot then
                                        local BarbarianLocationx = WorkedPlot:GetX()
                                        local BarbarianLocationy = WorkedPlot:GetY()
                                        if WorkedPlot:IsWater() then
                                            LuaEvents.LittleRevoltwater(Barbarianid,BarbarianLocationx,BarbarianLocationy)
                                        else
                                            LuaEvents.LittleRevolt(Barbarianid,BarbarianLocationx,BarbarianLocationy)   -- spawn barbarian
                                        end

                                    end
                                end
                            end
                        end
                        local OriginalOwner =    pCity:GetOriginalOwner()
                        local pPlayerCapital =   pPlayer:GetCities():GetCapitalCity()
                        local DistancebtweenCities = Map.GetPlotDistance(pCity:GetX(), pCity:GetY(), pPlayerCapital:GetX(), pPlayerCapital:GetY())
                        

                        if pPlayerCapital ~= pCity and not pCity:IsOccupied() and (OriginalOwner ~= player or DistancebtweenCities >= 30) then    -- Revolt Type 1
                            LuaEvents.RevoltType1(CityName,player)
                         
                        
                        elseif ReligionMOdValue < -50 then                                                                                        -- Revolt Typre 2 
                            local MajorReligion = pCity:GetReligion():GetMajorityReligion()
                            local MajorReligionName = Game.GetReligion():GetName(MajorReligion)
                            if MajorReligionName ~= "" then
                                LuaEvents.RevoltType2(CityName,player)
                            else
                                
                            end
                        elseif playerwarexhaustion <= -50 then                                                                                     -- Revolt Type 3
                            -- no idea for the moment..
                            
                            
                            
                            
                            
                            
                            
                            
                        end
                        
                        
                        
                    end  
                end
            else
            LuaEvents.restoreLoyalty( player, CityName);   
            end
            if CityLastRevolt2 + 11 > Currentturn() then
                local GoldLoss = (CityLastRevolt2 - Currentturn() - 11 ) * pCityData.GoldPerTurn /10 
                local FaithLoss = (CityLastRevolt2 - Currentturn() - 11 ) * pCityData.FaithPerTurn /10 
                local ScienceLoss = (CityLastRevolt2 - Currentturn() - 11 ) * pCityData.SciencePerTurn /10 
                local CulturalLoss = (CityLastRevolt2 - Currentturn() - 11 ) * pCityData.CulturePerTurn /10 
                
                
                LuaEvents.NoMoreTax( GoldLoss, player);                                                  -- loss gold per turn if recent revolt
                if ReligionMOdValue < -50 then
                     LuaEvents.NoMoreFaith( FaithLoss, player);                                         -- loss faith per turn if recent ReligionRevolt
                end
                if playerwarexhaustion <= -50 then                                                             -- loss science and culture if recent WEP revolt
                      LuaEvents.NoMoreScience( ScienceLoss, player);                               -- all effect can be stacked at same time
                      LuaEvents.NoMoreCulture( CulturalLoss, player);  
                end
            end
        end
    end
end
 
GameEvents.PlayerTurnStarted.Add(   RevoltType    )

-- ===========================================================================
--	GovernmentMod
-- ===========================================================================

function SetGovernmentMod(playerid,value)                            -- send GovernmentMod value to Script and add GovernmentMod to playerUI
        
    
        GovernmentMod[playerid] = GovernmentMod[playerid] + value
    if GovernmentMod[playerid] <= -100 then                         -- GovernmentMod can't be less than -100 
        GovernmentMod[playerid] = -100
    elseif GovernmentMod[playerid] >= 100 then                        -- GovernmentMod can't be less more than 100
        GovernmentMod[playerid] = 100     
    end
    LuaEvents.ChangePlayerGovernmentMod(playerid, GovernmentMod[playerid]);            -- create lua event to link with my script (thx Gedemon)
    SaveTableToSlot(GovernmentMod,"GovernmentMod")
end

function SetGovernmentModTooltip(playerid)
    local tooltip = ""
    if GovernmentFeaturetootip == nil then
        GovernmentFeaturetootip = {}
    end
    local GovernmentFeaturetootip2 = GovernmentFeaturetootip[playerid]
    if GovernmentFeaturetootip2 ~= nil then
        tooltip = tooltip..GovernmentFeaturetootip2.."[NEWLINE][NEWLINE]"
    end
    if setUIGovernmentchangedturn == nil then
            setUIGovernmentchangedturn = {}
    end
    local lastEventTurn = setUIGovernmentchangedturn[playerid]
    if lastEventTurn and lastEventTurn +10 > Currentturn()    then
        tooltip = tooltip.."[NEWLINE]-50 Changed Government recently"
    end
    if setUIPolicyChangedturn == nil then
            setUIPolicyChangedturn = {}
    end
     local lastEventTurn = setUIPolicyChangedturn[playerid]  
    if lastEventTurn and lastEventTurn +10 > Currentturn()    then
        tooltip = tooltip.."[NEWLINE]-3 per policies, Changed Policies recently"
    end
    return tooltip;
end

function OnGovernmentChanged( playerID:number )                                            -- lose loyalty when changing goverment
        SetGovernmentMod(playerID,-50) 
        if setUIGovernmentchangedturn == nil then
            setUIGovernmentchangedturn = {}
        end
        setUIGovernmentchangedturn[playerID] = Currentturn() 
end

Events.GovernmentChanged.Add( OnGovernmentChanged );

function OnPolicyChanged2(PlayerID)                                                        -- lose loyalty for every policies in government when change policies
    SetGovernmentMod(PlayerID,-3) 
    if setUIPolicyChangedturn == nil then
        setUIPolicyChangedturn = {}
    end
    setUIPolicyChangedturn[PlayerID] = Currentturn() 
end

Events.GovernmentPolicyChanged.Add(         OnPolicyChanged2 );

function GovernmentFeature(PlayerID)                                                   -- Government feature depending on current goverment and era
    if GovernmentFeaturetootip == nil then
        GovernmentFeaturetootip = {}
    end
    local pPlayer = Players[PlayerID]
    local governmentId = pPlayer:GetCulture():GetCurrentGovernment()
    local pPlayerMajorReligion =  pPlayer:GetReligion():GetReligionInMajorityOfCities()
    local PlayerAtWarid = IsAtWarWithWho(PlayerID)                                      -- id of player at war with pPlayer
    local pPlayerAtWarMajorReligion = -5
    local dice = math.random(0,1000) 
    local dice2 = math.random(0,1000) 
    local tooltip = ""
    if PlayerAtWarid ~= nil then
        local pPlayerAtWar = Players[PlayerAtWarid]
        pPlayerAtWarMajorReligion = pPlayerAtWar:GetReligion():GetReligionInMajorityOfCities()
    end
    if governmentId ~= -1 then
		local kCurrentGovernment = GameInfo.Governments[governmentId];
		local kGovernmentName = kCurrentGovernment.Name;	
        local localGovernmentName = Locale.Lookup(kGovernmentName)
        local tooltip2 = "[ICON_Government] Government is "..localGovernmentName..":[NEWLINE][NEWLINE]Event during this turn:[NEWLINE]"
        if kGovernmentName == "LOC_GOVERNMENT_CHIEFDOM_NAME" then
            SetWarexhaustionScripts(PlayerID,1)                                                        -- ChiefDom  reduced WEP 1
            GovernmentReligionMultiplier[PlayerID] = 1.5                                               -- SetGovermentReligionMultiplier to 1.5
            if dice < 100 then                                                                         -- moyenne = 5.25 gain per turn
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 Your chief was admirable" 
            elseif dice < 750 then
                SetGovernmentMod(PlayerID,10)   
                tooltip = tooltip.."+10 Your chief is alive and rule" 
            elseif dice < 850 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your chief was lamentable" 
            elseif dice < 900 then
                SetGovernmentMod(PlayerID,-10) 
                tooltip = tooltip.."-10 Your chief died"    
            elseif dice < 950 then
                SetGovernmentMod(PlayerID,-15) 
                tooltip = tooltip.."-15 Your chief has been killed" 
            else 
                SetGovernmentMod(PlayerID,-20) 
                tooltip = tooltip.."-20 Your chief took power by force" 
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_AUTOCRACY_NAME" then                                     -- autocracy
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 100 then                                                                             -- moyenne = 2.6 gain per turn        
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 Your leader was admirable" 
            elseif dice < 750 then
                SetGovernmentMod(PlayerID,5)   
                tooltip = tooltip.."+5 Your leader is alive and rule" 
            elseif dice < 900 then
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 Your leader was lamentable"
            elseif dice < 940 then
                SetGovernmentMod(PlayerID,-10)   
                tooltip = tooltip.."-10 Your leader died and your citizens liked him"
            else
                SetGovernmentMod(PlayerID,5)   
                tooltip = tooltip.."+5 Your leader is dead but your citizens did not like him"
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_OLIGARCHY_NAME" then
            SetWarexhaustionScripts(PlayerID,0.1)                                                          -- oligarchy  reduced WEP 0.1 
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 100 then                                                                             -- moyenne = 3.05 gain per turn
                SetGovernmentMod(PlayerID,10)   
                tooltip = tooltip.."+10 Your leaders were admirable" 
            elseif dice < 450 then
                SetGovernmentMod(PlayerID,8)   
                tooltip = tooltip.."+8 Your leaders are alive and rule" 
            elseif dice < 600 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leaders were lamentable"
            elseif dice < 850 then
                SetGovernmentMod(PlayerID,-4)   
                tooltip = tooltip.."-4 one of Your leaders died and your citizens liked him"
            else
                SetGovernmentMod(PlayerID,4)   
                tooltip = tooltip.."+4 one of Your leaders died but your citizens did not like him"
            end
            
        elseif kGovernmentName == "LOC_GOVERNMENT_CLASSICAL_REPUBLIC_NAME" then                            -- classical republic
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 800 then                                                                             -- moyenne = 9 - 0.65 = 8.35 gain per turn
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 the election was not rigged[NEWLINE]" 
            else
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 the election was rigged[NEWLINE]" 
            end
            if dice2 < 475 then                                                                         
                SetGovernmentMod(PlayerID,5)   
                tooltip = tooltip.."+5 Your leader was admirable" 
            elseif dice2 < 950 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader was lamentable" 
            elseif dice2 < 990 then
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-10 Your leader was killed while he was governing" 
            else
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader died while he was governing" 
            end
            
        elseif kGovernmentName == "LOC_GOVERNMENT_MONARCHY_NAME" then
             if pPlayerAtWarMajorReligion ~= pPlayerMajorReligion then                                  -- Monarchy if religion is different then reduced WEP 0.3
                SetWarexhaustionScripts(PlayerID,0.3)
            end
            GovernmentReligionMultiplier[PlayerID] = 1.5                                                 -- SetGovermentReligionMultiplier to 1.5
            if dice < 90 then                                                                            -- moyenne = 4.63  gain per turn
                SetGovernmentMod(PlayerID,20)   
                tooltip = tooltip.."+20 Your King was admirable" 
            elseif dice < 100 then
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 Your Queen was admirable" 
            elseif dice < 730 then
                SetGovernmentMod(PlayerID,10)   
                tooltip = tooltip.."+10 Your King is alive and rule" 
            elseif dice < 800 then
                SetGovernmentMod(PlayerID,5)   
                tooltip = tooltip.."+5 Your Queen is alive and rule" 
            elseif dice < 890 then
                SetGovernmentMod(PlayerID,-10)   
                tooltip = tooltip.."-10 Your King was Lamentable" 
            elseif dice < 900 then
                SetGovernmentMod(PlayerID,-20)   
                tooltip = tooltip.."-20 Your Queen was Lamentable" 
            elseif dice < 950 then
                SetGovernmentMod(PlayerID,-25) 
                tooltip = tooltip.."-25 Your monarch died and had an heir"    
            elseif dice < 960 then
                SetGovernmentMod(PlayerID,-35) 
                tooltip = tooltip.."-35 Your monarch died but had no heir"  
            elseif dice < 994 then
                SetGovernmentMod(PlayerID,-30) 
                tooltip = tooltip.."-30 Your monarch has been killed and had an heir" 
            elseif dice <= 999 then
                SetGovernmentMod(PlayerID,-40) 
                tooltip = tooltip.."-40 Your monarch has been killed but had no heir" 
            else 
                SetGovernmentMod(PlayerID,-50) 
                tooltip = tooltip.."-50 Your monarch took power by force or has no legitimacy" 
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_THEOCRACY_NAME" then
            if pPlayerAtWarMajorReligion ~= pPlayerMajorReligion then                                      -- theocracy if religion is different then reduced WEP 0.5
                SetWarexhaustionScripts(PlayerID,0.5)
            end
            GovernmentReligionMultiplier[PlayerID] = 2                                                     -- SetGovermentReligionMultiplier to 2
            if dice < 100 then                                                                             -- moyenne = 5.05 gain per turn 
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 Your leader was admirable" 
            elseif dice < 650 then
                SetGovernmentMod(PlayerID,9)   
                tooltip = tooltip.."+9 Your leader is alive and rule" 
            elseif dice < 800 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader was lamentable"
            elseif dice < 980 then
                SetGovernmentMod(PlayerID,-30)   
                tooltip = tooltip.."-30 Your leader died"
            else
                SetGovernmentMod(PlayerID,-40)   
                tooltip = tooltip.."-40 Your leader has been killed"
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_MERCHANT_REPUBLIC_NAME" then
            if pPlayerAtWarMajorReligion ~= pPlayerMajorReligion then                                      -- Merchant republic if religion is different then reduced WEP 0.1
                SetWarexhaustionScripts(PlayerID,0.1)
            end
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 800 then                                                                             -- moyenne = 9 - 0.65 = 8.35 gain per turn
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 the election was not rigged[NEWLINE]" 
            else
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 the election was rigged[NEWLINE]" 
            end
            if dice2 < 475 then                                                                         
                SetGovernmentMod(PlayerID,5)   
                tooltip = tooltip.."+5 Your leader was admirable" 
            elseif dice2 < 950 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader was lamentable" 
            elseif dice2 < 990 then
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-10 Your leader was killed while he was governing" 
            else
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader died while he was governing" 
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_FASCISM_NAME" then                                 
            SetWarexhaustionScripts(PlayerID,0.5)                                                          -- facism reduced WEP 0.5
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 200 then                                                                             -- moyenne = 2.575 gain per turn
                SetGovernmentMod(PlayerID,20)   
                tooltip = tooltip.."+20 Your leader was admirable" 
            elseif dice < 750 then
                SetGovernmentMod(PlayerID,3)   
                tooltip = tooltip.."+3 Your leader is alive and rule" 
            elseif dice < 900 then
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 Your leader was lamentable"
            elseif dice < 935 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader died"
            else
                SetGovernmentMod(PlayerID,-10)   
                tooltip = tooltip.."-10 Your leader has been killed"
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_COMMUNISM_NAME" then                                     -- communism
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 200 then                                                                             -- moyenne = 2.575 gain per turn
                SetGovernmentMod(PlayerID,20)   
                tooltip = tooltip.."+20 Your leader was admirable" 
            elseif dice < 750 then
                SetGovernmentMod(PlayerID,3)   
                tooltip = tooltip.."+3 Your leader is alive and rule" 
            elseif dice < 900 then
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 Your leader was lamentable"
            elseif dice < 935 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader died"
            else
                SetGovernmentMod(PlayerID,-10)   
                tooltip = tooltip.."-10 Your leader has been killed"
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_DEMOCRACY_NAME" then                                     -- democracy
            GovernmentReligionMultiplier[PlayerID] = 1                                                     -- SetGovermentReligionMultiplier to 1
            if dice < 950 then                                                                             -- moyenne = 13.5 - 0.295 = 13.205 gain per turn
                SetGovernmentMod(PlayerID,15)   
                tooltip = tooltip.."+15 the election was not rigged[NEWLINE]" 
            else
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 the election was rigged[NEWLINE]" 
            end
            if dice2 < 490 then                                                                         
                SetGovernmentMod(PlayerID,5)   
                tooltip = tooltip.."+5 Your leader was admirable" 
            elseif dice2 < 980 then
                SetGovernmentMod(PlayerID,-5)   
                tooltip = tooltip.."-5 Your leader was lamentable" 
            elseif dice2 <= 999 then
                SetGovernmentMod(PlayerID,-15)   
                tooltip = tooltip.."-15 Your leader was killed while he was governing" 
            else
                SetGovernmentMod(PlayerID,-10)   
                tooltip = tooltip.."-10 Your leader died while he was governing" 
            end
        elseif kGovernmentName == "LOC_GOVERNMENT_ANARCHY_NAME" then                                         -- anarchy
            GovernmentReligionMultiplier[PlayerID] = 1.5                                                     -- SetGovermentReligionMultiplier to 1.5
            SetGovernmentMod(PlayerID,-15)                                                                   -- moyenne = -15 loss per turn
            tooltip = tooltip.."-15 ANARCHY" 
        end
        if HistoryTooltip == nil then
           HistoryTooltip = {} 
        end
        local turn = Currentturn() 
        local khistoryKey = PlayerID.."."..turn
        HistoryTooltip[khistoryKey] = tooltip
        local tooltip = tooltip.."[NEWLINE][NEWLINE]Events of the last 5 Turns:"
        for i = turn -1 ,turn -5,-1 do
            local khistoryKey = PlayerID.."."..i
            if HistoryTooltip[khistoryKey] ~= nil then
                tooltip = tooltip.."[NEWLINE]"..HistoryTooltip[khistoryKey] 
            end
        end
        
        
        tooltip = tooltip2..tooltip
        GovernmentFeaturetootip[PlayerID] = tooltip
	end
    LuaEvents.ChangeLoyaltyGovernmentMod(PlayerID);
end

GameEvents.PlayerTurnStarted.Add( GovernmentFeature )

-- ===========================================================================
--	CONSTANTS
-- ===========================================================================

local ANIM_SPEED_RELIGION_CHANGE			:number = 1;
local COLOR_CITY_GREEN						:number	= 0xFF4CE710;
local COLOR_CITY_RED						:number	= 0xFF0101F5;
local COLOR_CITY_YELLOW						:number	= 0xFF2DFFF8;
local COLOR_HOLY_SITE						:number = 0xFFFFFFFF;
local COLOR_NO_MAJOR_RELIGION				:number = 0x00000000;
local COLOR_RELIGION_DEFAULT				:number = 0x02000000;

-- LOYALTY
local PRESSURE_BREAKDOWN_TYPE_POPULATION_PRESSURE	:string = "PopulationPressure";
local PRESSURE_BREAKDOWN_TYPE_GOVERNORS				:string = "Governors";
local PRESSURE_BREAKDOWN_TYPE_HAPPINESS				:string = "Happiness";
local PRESSURE_BREAKDOWN_TYPE_OTHER					:string = "Other";
local PRESSURE_BREAKDOWN_TYPE_CITY_STATE_BONUS		:string = "CityStateBonus";
local PRESSURE_BREAKDOWN_TYPE_FREE_CITY_BONUS		:string = "FreeCityBonus";

-- RELIGION
local DATA_FIELD_RELIGION_FOLLOWERS_IM		:string = "m_FollowersIM";
local DATA_FIELD_RELIGION_METERS_IM			:string = "m_MetersIM";
local DATA_FIELD_RELIGION_PRESSURE_CHANGES	:string = "m_PressureChanges";
local DATA_FIELD_RELIGION_PREV_FILL_PERCENT	:string = "m_FillPercent";
local DATA_FIELD_RELIGION_ICONS_IM			:string = "m_IconsIM";
local DATA_FIELD_RELIGION_FOLLOWER_LIST_IM	:string = "m_FollowerListIM";
local DATA_FIELD_RELIGION_POP_CHART_IM		:string = "m_PopChartIM";
local DATA_FIELD_PRODUCTION_CLICK_CALLBACK	:string = "m_ProductionClickCallback";

local RELIGION_POP_CHART_TOOLTIP_HEADER		:string = Locale.Lookup("LOC_CITY_BANNER_FOLLOWER_PRESSURE_TOOLTIP_HEADER");

local ICON_HOLY_SITE			:string = "Faith";
local ICON_PRESSURE_DOWN		:string = "PressureDown";
local ICON_PRESSURE_UP			:string = "PressureUp";
local MINIMUM_BANNER_WIDTH		:number = 186;
local PLOT_HIDDEN				:number	= 0;
local PLOT_REVEALED				:number	= 1;
local PLOT_VISIBLE				:number	= 2;
local PRESSURE_THRESHOLD_HIGH	:number = 400;
local PRESSURE_THRESHOLD_MEDIUM	:number = 200;
local PADDING_FOLLOWERS_BG		:number = 0;
local RELIGION_PRESSURE			:table = {
	NONE	= 0,
	LOW		= 1,
	MEDIUM	= 2,
	HIGH	= 3
};
local SIZE_HOLY_SITE_ICON		:number = 22;
local SIZE_RELIGION_ICON_LARGE	:number = 100;
local SIZE_RELIGION_ICON_SMALL	:number = 22;
local ZOOM_MULT_DELTA			:number = .01;
local ALPHA_DIM					:number = 0.45;

local YOFFSET_2DVIEW				:number = 26;
local ZOFFSET_3DVIEW				:number = 36;
local SIZEOFPOPANDPROD				:number = 80;	--The amount to add to the city banner to account for the size of the production icon and population number
local SIZEOFPOPANDPRODMETERS		:number = 15;	--The amount to add to the city banner backing width to allow for the production and population meters to appear

local BANNERTYPE_CITY_CENTER		:number = 0;
local BANNERTYPE_ENCAMPMENT			:number = 1;
local BANNERTYPE_AERODROME			:number = 2;
local BANNERTYPE_MISSILE_SILO		:number = 3;

local BANNERSTYLE_LOCAL_TEAM		:number = 0;
local BANNERSTYLE_OTHER_TEAM		:number = 1;

local m_pDirtyCityComponents	:table = {};
local m_preligionInfoance		:table = nil; -- tracks the most recently opened religion detail panel
local m_isReligionLensActive	:boolean = false;
local m_isLoyaltyLensActive		:boolean = false;
local m_isTradeSelectionActive	:boolean = false;

local m_refreshLocalPlayerRangeStrike:boolean = false;

-- ===========================================================================
--	MEMBERS
-- ===========================================================================
local m_zoomMultiplier		:number = 1;
local m_prevZoomMultiplier	:number = 1;

local CityBanner			:table = {};
local CityBannerInstances	:table = {};
local MiniBannerInstances	:table = {};

local m_CityBannerIM		:table = InstanceManager:new( "CityBanner",			"Anchor", Controls.CityBanners );
local m_AerodromeBannerIM	:table = InstanceManager:new( "AerodromeBanner",	"Anchor", Controls.CityBanners );
local m_WMDBannerIM			:table = InstanceManager:new( "WMDBanner",			"Anchor", Controls.CityBanners );
local m_EncampmentBannerIM	:table = InstanceManager:new( "EncampmentBanner",	"Anchor", Controls.CityBanners );
local m_HolySiteIconsIM		:table = InstanceManager:new( "HolySiteIcon",		"Anchor", Controls.CityDistrictIcons );

-- ===========================================================================
--	FUNCTIONS
-- ===========================================================================

-- ===========================================================================
--	Each city has a component ID that is internally 64-bits. 
--	The cityID is the lower 32-bits and will likely be the same across players
--	so both the playerID and cityID need to be used together in order to
--	obtain the proper city.
-- ===========================================================================
function GetCityBanner( playerID:number, cityID:number )
	if (CityBannerInstances[playerID] == nil) then
		return;
	end
	return CityBannerInstances[playerID][cityID];
end
-- ===========================================================================
function GetMiniBanner( playerID:number, districtID:number )	
	if (MiniBannerInstances[playerID] == nil) then
		return;
	end	
	return MiniBannerInstances[playerID][districtID];
end

-- ===========================================================================
-- constructor
-- ===========================================================================
function CityBanner:new( playerID: number, cityID : number, districtID : number, bannerType : number, bannerStyle : number )
	self = LuaClass.new(CityBanner);
	
	self.m_eMajorityReligion = -1; -- << Assign default values
	self.m_eLoyaltyWarningPlayer = -1; -- which player the city might flip to within 20 turns

	if bannerStyle == nil then UI.DataError("Missing bannerStyle: "..tostring(playerID)..", "..tostring(cityID)..", "..tostring(districtID)..", "..tostring(bannerType) ); end

	self:Initialize(playerID, cityID, districtID, bannerType, bannerStyle);

	if (bannerType == BANNERTYPE_CITY_CENTER) then
		if (CityBannerInstances[playerID] == nil) then
			CityBannerInstances[playerID] = {};
		end
		CityBannerInstances[playerID][cityID] = self;
	else
		if (MiniBannerInstances[playerID] == nil) then
			MiniBannerInstances[playerID] = {};
		end
		MiniBannerInstances[playerID][districtID] = self;
	end

	return self;
end

-- ===========================================================================
function CityBanner:destroy()
	if self.m_DetailStatusIM then
		self.m_DetailStatusIM:DestroyInstances();
	end
	if self.m_DetailEffectsIM then
		self.m_DetailEffectsIM:DestroyInstances();
	end
	if self.m_InfoIconIM then
		self.m_InfoIconIM:DestroyInstances();
	end
	if self.m_InfoConditionIM then
		self.m_InfoConditionIM:DestroyInstances();
	end
	if self.m_StatGovernorIM then
		self.m_StatGovernorIM:DestroyInstances();
	end
	if self.m_StatPopulationIM then
		self.m_StatPopulationIM:DestroyInstances();
	end
	if self.m_StatProductionIM then
		self.m_StatProductionIM:DestroyInstances();
	end
	if self.m_LoyaltyBreakdownIM then
		self.m_LoyaltyBreakdownIM:DestroyInstances();
	end
	if self.m_eLoyaltyWarningPlayer ~= -1 then
		local plot = Map.GetPlotIndex(self.m_PlotX, self.m_PlotY);
		UILens.ClearHex(LensLayers.LOYALTY_FREECITY_WARNING, plot);
	end
	
	if self.m_InstanceManager then
		self:UpdateSelected( false );
		if self.m_Instance then
			self.m_InstanceManager:ReleaseInstance( self.m_Instance );
		end
	end
end


-- ===========================================================================
function CityBanner:Initialize( playerID: number, cityID : number, districtID : number, bannerType : number, bannerStyle : number)

	self.m_Player = Players[playerID];
	self.m_DistrictID = districtID;
	self.m_CityID = cityID;

	self.m_Type = bannerType;
	self.m_Style = bannerStyle;
	self.m_IsSelected = false;
	self.m_IsCurrentlyVisible = false;
	self.m_IsForceHide = false;
	self.m_IsDimmed = false;
	self.m_OverrideDim = false;
	self.m_FogState = 0;
	self.m_UnitListEnabled = false;

	if (bannerType == BANNERTYPE_CITY_CENTER) then
		local pCity = self:GetCity();
		if pCity ~= nil then
			self.m_PlotX = pCity:GetX();
			self.m_PlotY = pCity:GetY();
		end

		-- Instantiate the banner
		self.m_InstanceManager = m_CityBannerIM;
		self.m_Instance = self.m_InstanceManager:GetInstance();

		-- Only create instance managers once, otherwise we can leak instances
		if self.m_DetailStatusIM == nil then
			self.m_DetailStatusIM = InstanceManager:new( "CityDetailStatus", "Icon", self.m_Instance.CityDetailsStatus );
		end
		if self.m_DetailEffectsIM == nil then
			self.m_DetailEffectsIM = InstanceManager:new( "CityDetailEffect", "Icon", self.m_Instance.CityDetailsEffects );
		end
		if self.m_InfoIconIM == nil then
			self.m_InfoIconIM = InstanceManager:new( "CityInfoType", "Button", self.m_Instance.CityInfoStack );
		end
		if self.m_InfoConditionIM == nil then
			self.m_InfoConditionIM = InstanceManager:new( "CityInfoCondition", "Button", self.m_Instance.CityInfoStack );
		end
		if self.m_StatGovernorIM == nil then
			self.m_StatGovernorIM = InstanceManager:new( "CityStatGovernor", "Button", self.m_Instance.CityStatusStack );
		end
		if self.m_StatProductionIM == nil then
			self.m_StatProductionIM = InstanceManager:new( "CityStatProduction", "Button", self.m_Instance.CityStatusStack );
		end
		
		-- If instance managers need to be re-created, you can clean up instances manually
		if self.m_StatPopulationIM then
			self.m_StatPopulationIM:DestroyInstances();
			self.m_StatPopulationIM = nil;
		end

		self.m_Instance.CityBannerButton:RegisterCallback( Mouse.eLClick, OnCityBannerClick );
		self.m_Instance.CityBannerButton:SetVoid1(playerID);
		self.m_Instance.CityBannerButton:SetVoid2(cityID);

		if (bannerStyle == BANNERSTYLE_LOCAL_TEAM and playerID == Game.GetLocalPlayer()) then
			self.m_StatPopulationIM = InstanceManager:new( "CityStatPopulation", "Button", self.m_Instance.CityStatusStack );
			self.m_Instance.CityStrikeButton:RegisterCallback( Mouse.eLClick, OnCityStrikeButtonClick );
			self.m_Instance.CityStrikeButton:SetVoid1(playerID);
			self.m_Instance.CityStrikeButton:SetVoid2(cityID);
		end

		-- If we're not local player, show limited population info
		if self.m_StatPopulationIM == nil then
			self.m_StatPopulationIM	= InstanceManager:new( "CityStatPopulationLimited", "BG", self.m_Instance.CityStatusStack );
		end

		self:UpdateReligion();

		local loyaltyInfo:table = self.m_Instance.LoyaltyInfo;
		local toggleLoyalty = function(on)
			return function()
				loyaltyInfo.CulturalIdentityButton:SetHide(on);
				loyaltyInfo.CulturalIdentityExpandedButton:SetHide(not on);
			end
		end
		loyaltyInfo.CulturalIdentityButton:RegisterCallback( Mouse.eLClick, toggleLoyalty(true) );
		loyaltyInfo.CulturalIdentityExpandedButton:RegisterCallback( Mouse.eLClick, toggleLoyalty(false) );

		if self.m_LoyaltyBreakdownIM == nil then
			self.m_LoyaltyBreakdownIM = InstanceManager:new("InfluenceLineInstance","Top", self.m_Instance.LoyaltyInfo.IdentityBreakdownStack);
		end

		self:UpdateLoyalty();
	elseif (bannerType == BANNERTYPE_AERODROME) then
		self:CreateAerodromeBanner();
		self:UpdateAerodromeBanner();
	elseif (bannerType == BANNERTYPE_MISSILE_SILO) then
		self:CreateWMDBanner();
		self:UpdateWMDBanner();
	elseif (bannerType == BANNERTYPE_ENCAMPMENT) then
		self:CreateEncampmentBanner();
		self:UpdateEncampmentBanner();
	end

	self:UpdateName();
	self:UpdateStats();
	self:UpdatePosition();
	self:UpdateVisibility();
	self:UpdateRangeStrike();
	self:UpdateColor();
end

-- ===========================================================================
function CityBanner:CreateAerodromeBanner()
	-- Set the appropriate instance factory (mini banner one) for this flag...
	self.m_InstanceManager = m_AerodromeBannerIM;
	self.m_Instance = self.m_InstanceManager:GetInstance();

	self.m_IsImprovementBanner = false;

	local pDistrict = self:GetDistrict();
	if (pDistrict ~= nil) then
		self.m_PlotX = pDistrict:GetX();
		self.m_PlotY = pDistrict:GetY();
	else	-- it's an banner not associated with a district, so the districtID should be a plot index
		self.m_PlotX, self.m_PlotY = Map.GetPlotLocation(self.m_DistrictID);
		self.m_IsImprovementBanner = true;
	end
end

-- ===========================================================================
function CityBanner:UpdateAerodromeBanner()
	self.m_Instance.UnitListPopup:ClearEntries();
	
	local iAirCapacity = 0;
	local iAirUnitCount = 0;

	local pDistrict : table = self:GetDistrict();
	if (pDistrict ~= nil) then
		-- Update minibanner for aerodrome
		iAirCapacity = pDistrict:GetAirSlots();
		local bHasAirUnits, tAirUnits = pDistrict:GetAirUnits();
		if (bHasAirUnits and tAirUnits ~= nil) then
			-- Update unit instances in unit list
			for i,unit in ipairs(tAirUnits) do
				local unitEntry:table = {};
				self.m_Instance.UnitListPopup:BuildEntry( "UnitListEntry", unitEntry );

				-- Update name
				unitEntry.UnitName:SetText( Locale.ToUpper( unit:GetName() ) );

				-- Update icon
				local iconInfo:table = GetUnitIcon(unit, 22);
				if iconInfo.textureSheet then
					unitEntry.UnitTypeIcon:SetTexture( iconInfo.textureOffsetX, iconInfo.textureOffsetY, iconInfo.textureSheet );
				end

				-- Update callback
				unitEntry.Button:RegisterCallback( Mouse.eLClick, OnUnitSelected );
				unitEntry.Button:SetVoid1(playerID);
				unitEntry.Button:SetVoid2(unit:GetID());

				-- Increment count
				iAirUnitCount = iAirUnitCount + 1;

				-- Fade out the button icon and text if the unit is not able to move
				if unit:IsReadyToMove() then
					unitEntry.UnitName:SetAlpha(1.0);
					unitEntry.UnitTypeIcon:SetAlpha(1.0);
				else
					unitEntry.UnitName:SetAlpha(ALPHA_DIM);
					unitEntry.UnitTypeIcon:SetAlpha(ALPHA_DIM);
				end
			end
		end
	else
		-- Update minibanner for airstrip
		local airstripPlot = Map.GetPlotByIndex(self.m_DistrictID);
		local tAirUnits = airstripPlot:GetAirUnits();
		if tAirUnits then
			local eImprovement = airstripPlot:GetImprovementType();
			if (eImprovement ~= -1) then
				iAirCapacity = GameInfo.Improvements[eImprovement].AirSlots;
			end

			-- Update unit instances in unit list
			for i,unit in ipairs(tAirUnits) do
				local unitEntry:table = {};
				self.m_Instance.UnitListPopup:BuildEntry( "UnitListEntry", unitEntry );

				-- Update name
				unitEntry.UnitName:SetText( Locale.ToUpper(unit:GetName()) );

				-- Update icon
				local iconInfo:table = GetUnitIcon(unit, 22, true);
				if iconInfo.textureSheet then
					unitEntry.UnitTypeIcon:SetTexture( iconInfo.textureOffsetX, iconInfo.textureOffsetY, iconInfo.textureSheet );
				end

				-- Update callback
				unitEntry.Button:RegisterCallback( Mouse.eLClick, OnUnitSelected );
				unitEntry.Button:SetVoid1(playerID);
				unitEntry.Button:SetVoid2(unit:GetID());

				-- Increment count
				iAirUnitCount = iAirUnitCount + 1;

				-- Fade out the button icon and text if the unit is not able to move
				if unit:IsReadyToMove() then
					unitEntry.UnitName:SetAlpha(1.0);
					unitEntry.UnitTypeIcon:SetAlpha(1.0);
				else
					unitEntry.UnitName:SetAlpha(ALPHA_DIM);
					unitEntry.UnitTypeIcon:SetAlpha(ALPHA_DIM);
				end
			end
		end
	end

	-- Update current and max air unit capacity
	self.m_Instance.AerodromeCurrentUnitCount:SetText(iAirUnitCount);
	self.m_Instance.AerodromeMaxUnitCount:SetText(iAirCapacity);

	-- Update tooltip to show unit capacity
	self.m_Instance.AerodromeBase:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_AERODROME_AIRCRAFT_STATIONED", iAirUnitCount, iAirCapacity));

	-- If current air unit count is 0 then disabled popup
	if iAirUnitCount <= 0 then
		self.m_UnitListEnabled = false;
	else
		self.m_UnitListEnabled = true;
	end

	self:SetFogState( self.m_FogState );

	self.m_Instance.UnitListPopup:CalculateInternals();

	-- Adjust the scroll panel offset so stack is centered whether scrollbar is visible or not
	local scrollPanel = self.m_Instance.UnitListPopup:GetScrollPanel();
	if scrollPanel then
		if scrollPanel:GetScrollBar():IsHidden() then
			scrollPanel:SetOffsetX(0);
		else
			scrollPanel:SetOffsetX(7);
		end
	end
		
	self.m_Instance.UnitListPopup:ReprocessAnchoring();
	self.m_Instance.UnitListPopup:GetGrid():ReprocessAnchoring();
end

-- ===========================================================================
function CityBanner:CreateWMDBanner()
	-- Set the appropriate instance factory (mini banner one) for this flag...
	self.m_InstanceManager = m_WMDBannerIM;
	self.m_Instance = self.m_InstanceManager:GetInstance();

	self.m_IsImprovementBanner = false;

	local pDistrict = self:GetDistrict();
	if (pDistrict ~= nil) then
		self.m_PlotX = pDistrict:GetX();
		self.m_PlotY = pDistrict:GetY();
	else	-- it's an banner not associated with a district, so the districtID should be a plot index
		self.m_PlotX, self.m_PlotY = Map.GetPlotLocation(self.m_DistrictID);
		self.m_IsImprovementBanner = true;
	end

	-- Setup button callbacks
	local plotID = Map.GetPlotIndex(self.m_PlotX, self.m_PlotY);
	local eNuclearDevice = GameInfo.WMDs["WMD_NUCLEAR_DEVICE"].Index;
	self.m_Instance.NukeBombButton:RegisterCallback( Mouse.eLClick, OnICBMStrikeButtonClick );
	self.m_Instance.NukeBombButton:SetVoid1(plotID);
	self.m_Instance.NukeBombButton:SetVoid2(eNuclearDevice);

	local eThermonuclearDevice = GameInfo.WMDs["WMD_THERMONUCLEAR_DEVICE"].Index;
	self.m_Instance.ThermoNukeBombButton:RegisterCallback( Mouse.eLClick, OnICBMStrikeButtonClick );
	self.m_Instance.ThermoNukeBombButton:SetVoid1(plotID);
	self.m_Instance.ThermoNukeBombButton:SetVoid2(eThermonuclearDevice);
end

-- ===========================================================================
function CityBanner:UpdateWMDBanner()
	
	local pCity:table = self:GetCity();

	-- Don't show the mini banner if this silo doesn't belong to the local player
	if pCity ~= nil and pCity:GetOwner() ~= Game.GetLocalPlayer() then
		self.m_Instance.WMDBannerContainer:SetHide(true);
		return;
	end
	self.m_Instance.WMDBannerContainer:SetHide(false);

	local playerWMDs = self.m_Player:GetWMDs();

	for entry in GameInfo.WMDs() do
		if (entry.WeaponType == "WMD_NUCLEAR_DEVICE") then
			local count = playerWMDs:GetWeaponCount(entry.Index);
			if (count > 0) then
				-- Player has nukes
				self.m_Instance.NukeCountLabel:SetText(count);
				self.m_Instance.NukeBombButtonBackground:SetHide(false);

				-- Check if we're able to fire
				local bSiloCanFire:boolean = false;
				if( pCity ~= nil ) then
					local tParameters = {};
					tParameters[CityCommandTypes.PARAM_WMD_TYPE] = entry.Index;
					tParameters[CityCommandTypes.PARAM_X0] = self.m_PlotX;
					tParameters[CityCommandTypes.PARAM_Y0] = self.m_PlotY;
					local tResults = CityManager.GetCommandTargets(pCity, CityCommandTypes.WMD_STRIKE, tParameters);
					local allPlots = tResults[CityCommandResults.PLOTS];
					if (allPlots ~= nil) then
						bSiloCanFire = true;
					end
				end

				-- Update button state and tooltip
				if( bSiloCanFire) then
					self.m_Instance.NukeBombButton:SetDisabled(false);
					self.m_Instance.NukeBombButton:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_NUCLEAR_STRIKE_CAPABLE"));
				else
					self.m_Instance.NukeBombButton:SetDisabled(true);
					self.m_Instance.NukeBombButton:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_WEAPON_UNAVAILABLE"));
				end
			else
				-- Player does not have nukes
				self.m_Instance.NukeCountLabel:SetText("0");
				self.m_Instance.NukeBombButtonBackground:SetHide(true);
			end
		elseif (entry.WeaponType == "WMD_THERMONUCLEAR_DEVICE") then
			local count = playerWMDs:GetWeaponCount(entry.Index);
			if (count > 0) then
				-- Player has thermonuclear bombs
				self.m_Instance.ThermoNukeCountLabel:SetText(count);
				self.m_Instance.ThermoNukeBombButtonBackground:SetHide(false);

				-- Check if we're able to fire
				local bSiloCanFire:boolean = false;
				if( pCity ~= nil ) then
					local tParameters = {};
					tParameters[CityCommandTypes.PARAM_WMD_TYPE] = entry.Index;
					tParameters[CityCommandTypes.PARAM_X0] = self.m_PlotX;
					tParameters[CityCommandTypes.PARAM_Y0] = self.m_PlotY;
					local tResults = CityManager.GetCommandTargets(pCity, CityCommandTypes.WMD_STRIKE, tParameters);
					local allPlots = tResults[CityCommandResults.PLOTS];
					if (allPlots ~= nil) then
						bSiloCanFire = true;
					end
				end

				-- Update button state and tooltip
				if( bSiloCanFire) then
					self.m_Instance.ThermoNukeBombButton:SetDisabled(false);
					self.m_Instance.ThermoNukeBombButton:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_THERMONUCLEAR_STRIKE_CAPABLE"));
				else
					self.m_Instance.ThermoNukeBombButton:SetDisabled(true);
					self.m_Instance.ThermoNukeBombButton:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_WEAPON_UNAVAILABLE"));
				end
			else
				-- Player does not have thermonuclear bombs
				self.m_Instance.ThermoNukeCountLabel:SetText("0");
				self.m_Instance.ThermoNukeBombButtonBackground:SetHide(true);
			end
		end
	end
end

-- ===========================================================================
function CityBanner:CreateEncampmentBanner()
	-- Set the appropriate instance factory (mini banner one) for this flag...
	self.m_InstanceManager = m_EncampmentBannerIM;
	self.m_Instance = self.m_InstanceManager:GetInstance();

	self.m_IsImprovementBanner = false;

	local pDistrict = self:GetDistrict();
	if (pDistrict ~= nil) then
		self.m_PlotX = pDistrict:GetX();
		self.m_PlotY = pDistrict:GetY();

		-- Update district strength
		local districtDefense:number = math.floor(pDistrict:GetDefenseStrength() + 0.5);
		self.m_Instance.DistrictDefenseStrengthLabel:SetText(districtDefense);

		-- Setup strike button callback
		self.m_Instance.CityStrikeButton:RegisterCallback( Mouse.eLClick, OnDistrictRangeStrikeButtonClick );
		self.m_Instance.CityStrikeButton:SetVoid1(self.m_Player:GetID());
		self.m_Instance.CityStrikeButton:SetVoid2(self.m_DistrictID);
	end
end

-- ===========================================================================
function CityBanner:UpdateEncampmentBanner()
	-- Update wall/district health
	local pDistrict:table = self:GetDistrict();

	local districtDefense		:number = math.floor(pDistrict:GetDefenseStrength() + 0.5);
	local districtHitpoints		:number = pDistrict:GetMaxDamage(DefenseTypes.DISTRICT_GARRISON);
	local currentDistrictDamage :number = pDistrict:GetDamage(DefenseTypes.DISTRICT_GARRISON);
	local wallHitpoints			:number = pDistrict:GetMaxDamage(DefenseTypes.DISTRICT_OUTER);
	local currentWallDamage		:number = pDistrict:GetDamage(DefenseTypes.DISTRICT_OUTER);
	local healthTooltip			:string = Locale.Lookup("LOC_CITY_BANNER_DISTRICT_HITPOINTS", ((districtHitpoints-currentDistrictDamage) .. "/" .. districtHitpoints));	
	local defTooltip			:string = Locale.Lookup("LOC_CITY_BANNER_DISTRICT_DEFENSE_STRENGTH", districtDefense);

	if (wallHitpoints > 0) then
		self.m_Instance.CityDefenseBar:SetHide(false);
		healthTooltip = healthTooltip .. "[NEWLINE]" .. Locale.Lookup("LOC_CITY_BANNER_OUTER_DEFENSE_HITPOINTS", ((wallHitpoints-currentWallDamage) .. "/" .. wallHitpoints));
		self.m_Instance.CityDefenseBar:SetPercent((wallHitpoints-currentWallDamage) / wallHitpoints);
		self:SetHealthBarColor();
	else
		self.m_Instance.CityDefenseBar:SetHide(true);
	end

	if districtHitpoints < 0 or (((districtHitpoints-currentDistrictDamage) / districtHitpoints) == 1 and wallHitpoints == 0) then
		self.m_Instance.CityHealthBar:SetHide(true);
	else
		self.m_Instance.CityHealthBar:SetHide(false);
		self.m_Instance.CityHealthBar:SetPercent((districtHitpoints-currentDistrictDamage) / districtHitpoints);	
	end

	self.m_Instance.EncampmentBannerContainer:SetToolTipString(healthTooltip);
	self.m_Instance.DistrictDefenseGrid:SetToolTipString(defTooltip);
end

-- ===========================================================================
function OnUnitSelected( playerID:number, unitID:number )
	if Game.GetLocalPlayer() >= 0 then
		local playerUnits:table = Players[Game.GetLocalPlayer()]:GetUnits();
		if playerUnits then
			local selectedUnit:table = playerUnits:FindID(unitID);
			if selectedUnit then
				UI.SelectUnit( selectedUnit );
			end
		end
	end
end

-- ===========================================================================
function CityBanner:IsVisible()
	if Game.GetLocalPlayer() >= 0 then
		local pLocalPlayerVis:table = PlayersVisibility[Game.GetLocalPlayer()];
		local city:table = self:GetCity();
		local locX:number = city:GetX();
		local locY:number = city:GetY();
		if pLocalPlayerVis:IsVisible(locX, locY) then
			return true;
		elseif pLocalPlayerVis:IsRevealed(locX, locY) then
			return true;
		end
	end
	return false;
end

function CityBanner:IsTeam()
	return self.m_Style == BANNERSTYLE_LOCAL_TEAM;
end

-- ===========================================================================
-- Resize and recenter city banner images to accomodate the city name
function CityBanner:Resize()
	if (self.m_Type == BANNERTYPE_CITY_CENTER) then
		local pCity : table = self:GetCity();
		if (pCity ~= nil) then
			self.m_Instance.CityInfoStack:CalculateSize();
			self.m_Instance.CityStatusStack:CalculateSize();
			local nameContainerSize = self.m_Instance.CityName:GetSizeX();

			self.m_Instance.ContentStack:CalculateSize();
			local newBannerSize:number = self.m_Instance.ContentStack:GetSizeX() + 20;

			self.m_Instance.CityDetailsStack:CalculateSize();
			self.m_Instance.CityDetails:SetSizeX(self.m_Instance.CityDetailsStack:GetSizeX() + 10);

			local topInfoSize:number = self.m_Instance.CityDetails:GetSizeX() + 12;
			if (newBannerSize < topInfoSize) then
				newBannerSize = topInfoSize;
			end
			if (newBannerSize < MINIMUM_BANNER_WIDTH) then
				newBannerSize = MINIMUM_BANNER_WIDTH;
			end
			self.m_Instance.Container:SetSizeX(newBannerSize);

			-- Inside the city strength indicator (shield) - Recentering the characters that have odd leading
			--if(self.m_Instance.DefenseNumber:GetText() == "4" or self.m_Instance.DefenseNumber:GetText() == "6" or self.m_Instance.DefenseNumber:GetText() == "8") then
			--	self.m_Instance.DefenseNumber:SetOffsetX(-2);
			--end
		end
	end
end

-- ===========================================================================
-- Assign player colors to the appropriate banner elements
function CityBanner:UpdateColor()

	local backColor, frontColor  = UI.GetPlayerColors( self.m_Player:GetID() );
	local darkerBackColor = DarkenLightenColor(backColor,(-85),238);
	local brighterBackColor = DarkenLightenColor(backColor,90,255);

	if self.m_IsSelected then
		-- TODO: Highlight selected city
	end

	if (self.m_Type == BANNERTYPE_CITY_CENTER) then
		self.m_Instance.CityBannerFill:SetColor( backColor );
		self.m_Instance.CityBannerFill2:SetColor( darkerBackColor );
		self.m_Instance.CityBannerFill3:SetColor( brighterBackColor );
		self.m_Instance.CityBannerFillOver:SetColor( frontColor );
		self.m_Instance.CityBannerFillOut:SetColor( frontColor );
		self.m_Instance.CityName:SetColor( frontColor, 0 );
		self.m_Instance.CityName:SetColor( darkerBackColor, 1 );
		if self.m_CivIconInstance then
			self.m_CivIconInstance.Icon:SetColor( frontColor );
		end
	elseif (self.m_Type == BANNERTYPE_AERODROME) then
		if self.m_Instance.AerodromeUnitsButton_Base ~= nil then
			self.m_Instance.AerodromeUnitsButton_Base:SetColor( backColor );
			self.m_Instance.AerodromeUnitsButton_Darker:SetColor( darkerBackColor );
			self.m_Instance.AerodromeUnitsButton_Lighter:SetColor( brighterBackColor );
			self.m_Instance.AerodromeUnitsButton_None:SetColor( frontColor );
			self.m_Instance.AerodromeUnitsButtonIcon:SetColor( frontColor );
		end
	elseif (self.m_Type == BANNERTYPE_MISSILE_SILO) then
		if self.m_Instance.Banner_Base ~= nil then
			self.m_Instance.Banner_Base:SetColor( backColor );
			self.m_Instance.Banner_Darker:SetColor( darkerBackColor );
			self.m_Instance.Banner_Lighter:SetColor( brighterBackColor );
			self.m_Instance.Banner_None:SetColor( frontColor );
			self.m_Instance.NukeCountLabel:SetColor( frontColor );
			self.m_Instance.ThermoNukeCountLabel:SetColor( frontColor );
		end
	elseif (self.m_Type == BANNERTYPE_ENCAMPMENT) then
		if self.m_Instance.Banner_Base ~= nil then
			self.m_Instance.Banner_Base:SetColor( backColor );
			self.m_Instance.Banner_Darker:SetColor( darkerBackColor );
			self.m_Instance.Banner_Lighter:SetColor( brighterBackColor );
			self.m_Instance.Banner_None:SetColor( frontColor );
		end
	else
		self.m_Instance.MiniBannerBackground:SetColor( backColor );
	end

	self:SetHealthBarColor();

end

-- ===========================================================================
function CityBanner:SetHealthBarColor()
	if self.m_Instance.CityHealthBar == nil then
		-- This normal behaviour in the case of missile silo and aerodrome minibanners
		return;
	end

	local percent = self.m_Instance.CityHealthBar:GetPercent();
	if (percent > .8 ) then
		self.m_Instance.CityHealthBar:SetColor( COLOR_CITY_GREEN );
	elseif ( percent > .4) then
		self.m_Instance.CityHealthBar:SetColor( COLOR_CITY_YELLOW );
	elseif ( percent < .4) then
		self.m_Instance.CityHealthBar:SetColor( COLOR_CITY_RED ); 
	end
end

-- ===========================================================================
function OnPopulationIconClicked(playerID: number, cityID: number)
	OnCityBannerClick(playerID, cityID);
	if (playerID == Game.GetLocalPlayer()) then
		LuaEvents.CityPanel_ToggleManageCitizens();
	end
end

-- ===========================================================================
-- Non-instance function so it can be overwritten by mods
function CityBanner:UpdatePopulation(isLocalPlayer:boolean, pCity:table, pCityGrowth:table)

	self.m_StatPopulationIM:ResetInstances();

	local currentPopulation:number = pCity:GetPopulation();
	local populationInstance:table = self.m_StatPopulationIM:GetInstance();

	if isLocalPlayer then
		local food				:number = pCityGrowth:GetFood();
		local foodSurplus		:number = pCityGrowth:GetFoodSurplus();
		local isGrowing			:boolean= pCityGrowth:GetTurnsUntilGrowth() ~= -1;
		local isStarving		:boolean= pCityGrowth:GetTurnsUntilStarvation() ~= -1;
		local growthThreshold	:number = pCityGrowth:GetGrowthThreshold();
		local foodpct			:number = Clamp( food / growthThreshold, 0.0, 1.0 );

		local turnsUntilGrowth:number = 0;	-- It is possible for zero... no growth and no starving.
		if isGrowing then
			turnsUntilGrowth = pCityGrowth:GetTurnsUntilGrowth();
		elseif isStarving then
			turnsUntilGrowth = -pCityGrowth:GetTurnsUntilStarvation();	-- negative
		end
		
		local popTooltip:string = Locale.Lookup("LOC_CITY_BANNER_POPULATION") .. ": " .. currentPopulation;
		if turnsUntilGrowth > 0 then
			popTooltip = popTooltip .. "[NEWLINE]  " .. Locale.Lookup("LOC_CITY_BANNER_TURNS_GROWTH", turnsUntilGrowth);
			popTooltip = popTooltip .. "[NEWLINE]  " .. Locale.Lookup("LOC_CITY_BANNER_FOOD_SURPLUS", round(foodSurplus,1));
		elseif turnsUntilGrowth == 0 then
			popTooltip = popTooltip .. "[NEWLINE]  " .. Locale.Lookup("LOC_CITY_BANNER_STAGNATE");
		elseif turnsUntilGrowth < 0 then
			popTooltip = popTooltip .. "[NEWLINE]  " .. Locale.Lookup("LOC_CITY_BANNER_TURNS_STARVATION", -turnsUntilGrowth);
		end
		populationInstance.FillMeter:SetToolTipString(popTooltip);
		if turnsUntilGrowth ~= 0 then
			populationInstance.CityPopTurnsLeft:SetText(turnsUntilGrowth);
		else
			populationInstance.CityPopTurnsLeft:SetText("-");
		end

		populationInstance.FillMeter:SetPercent(foodpct);
		populationInstance.SlotMeter:SetPercent(1 - foodpct);

		populationInstance.Button:RegisterCallback(Mouse.eLClick, OnPopulationIconClicked);
		populationInstance.Button:SetVoid1(Game.GetLocalPlayer());
		populationInstance.Button:SetVoid2(pCity:GetID());
	else
		populationInstance.BG:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_POPULATION") .. ": " .. currentPopulation);
	end

	populationInstance.CityPopulation:SetText(tostring(currentPopulation));
end

-- ===========================================================================
function OnGovernorIconClicked(playerID: number, cityID: number)
	if (playerID == Game.GetLocalPlayer()) then
		OnCityBannerClick(playerID, cityID);
	else
		OnCityBannerLookAt(playerID, cityID);
	end
	LuaEvents.GovernorPanel_Toggle();
end

-- ===========================================================================
-- Non-instance function so it can be overwritten by mods
function CityBanner:UpdateGovernor(pCity:table)
	
	self.m_StatGovernorIM:ResetInstances();

	local localPlayerID:number = Game.GetLocalPlayer();

	if localPlayerID < 0 then
		return;
	end

	local pLocalPlayer:table = Players[localPlayerID];
	local pLocalPlayerDiplomacy:table = pLocalPlayer:GetDiplomacy();

	local isCityState:boolean = false;
	local cityOwner:number = pCity:GetOwner();
	for i, pCityState in ipairs(PlayerManager.GetAliveMinors()) do
		if pCityState:GetID() == cityOwner then
			isCityState = true;
			break;
		end
	end

	-- Always show local players governor first
	local otherGovernors = 0;
	local otherGovernorsTT = Locale.Lookup("LOC_CITY_STATE_PANEL_HAS_AMBASSADOR_TOOLTIP");

	local governors:table = pCity:GetAllAssignedGovernors();
	table.sort(governors, function(a, b) return a:GetOwner() == localPlayerID end);

	for i, pGovernor in ipairs(governors) do
		local playerID:number = pGovernor:GetOwner();
		local visibility:number = DiplomaticVisibilityTypes.TOP_SECRET + 1;

		if playerID ~= localPlayerID then
			visibility = pLocalPlayerDiplomacy:GetVisibilityOn(playerID);
			if visibility <= DiplomaticVisibilityTypes.NONE then
				return; -- Cannot see governor in this city
			end
		end

		-- Always show full icon for local players governors, even on city states
		if not isCityState or playerID == localPlayerID then
			local instance:table = self.m_StatGovernorIM:GetInstance();
			instance.NumOfAmbassadors:SetText("");

			if visibility > DiplomaticVisibilityTypes.LIMITED then
				
				instance.UnknownGovernor:SetHide(true);
				local icon = "ICON_" .. GameInfo.Governors[pGovernor:GetType()].GovernorType;
				instance.SlotMeter:SetTexture(IconManager:FindIconAtlas(icon .. "_SLOT", 32));
				instance.FillMeter:SetTexture(IconManager:FindIconAtlas(icon .. "_FILL", 32));

				if visibility > DiplomaticVisibilityTypes.FULL then
					if (pGovernor:IsEstablished()) then
						instance.TurnsLeft:SetText("");
						instance.FillMeter:SetPercent(1);
						instance.SlotMeter:SetPercent(0);
						instance.FillMeter:SetToolTipString(Locale.Lookup("LOC_HUD_CITY_GOVERNOR_ESTABLISHED_SPECIFIC", pGovernor:GetName()));
					else
						local iTurnsOnSite:number = pGovernor:GetTurnsOnSite();
						local iTurnsToEstablish:number = pGovernor:GetTurnsToEstablish();
						local iTurnsUntilEstablished:number = iTurnsToEstablish - iTurnsOnSite;
						local establishedPct:number = (iTurnsToEstablish - iTurnsUntilEstablished) / iTurnsToEstablish;

						instance.TurnsLeft:SetText(tostring(iTurnsUntilEstablished));
						instance.FillMeter:SetPercent(establishedPct);
						instance.SlotMeter:SetPercent(1 - establishedPct);

						local tooltip:string = Locale.Lookup("LOC_HUD_CITY_GOVERNOR_ASSIGNED_SPECIFIC", pGovernor:GetName());
						tooltip = tooltip .. "[NEWLINE]";
						tooltip = tooltip .. Locale.Lookup("LOC_GOVERNORS_SCREEN_GOVERNOR_TURNS_UNTIL_ESTABLISHED", iTurnsUntilEstablished);
						instance.FillMeter:SetToolTipString(tooltip);
					end
				else -- Open visibility (can see governor, but not whether they are established)
					instance.Button:SetVisState(-1); -- This makes production icon hide
					instance.TurnsLeft:SetText("?");
					instance.FillMeter:SetPercent(0);
					instance.SlotMeter:SetPercent(1);
					instance.FillMeter:SetToolTipString(Locale.Lookup("LOC_HUD_CITY_GOVERNOR_NO_TURNS", pGovernor:GetName()));
				end
			else -- Limited visibility (can see there is a governor, but not who or whether they are established)
				instance.TurnsLeft:SetText("");
				instance.FillMeter:SetPercent(0);
				instance.SlotMeter:SetPercent(0);
				instance.UnknownGovernor:SetHide(false);
				instance.FillMeter:SetToolTipString(Locale.Lookup("LOC_HUD_CITY_GOVERNOR_NO_ID"));
			end
			
			instance.Button:RegisterCallback(Mouse.eLClick, OnGovernorIconClicked);
			instance.Button:SetVoid1(cityOwner);
			instance.Button:SetVoid2(pCity:GetID());

		else -- We must be a city state, multiple diplomats can be slotted here (but only show one icon)
			local playerID:number = pGovernor:GetOwner();
			local visibility:number = DiplomaticVisibilityTypes.TOP_SECRET + 1;

			if playerID ~= localPlayerID then
				otherGovernors = otherGovernors + 1;
				visibility = pLocalPlayerDiplomacy:GetVisibilityOn(playerID);
			end

			local establishInfo;
			if visibility > DiplomaticVisibilityTypes.FULL then
				if (pGovernor:IsEstablished()) then
					establishInfo = Locale.Lookup("LOC_HUD_CITY_GOVERNOR_ESTABLISHED");
				else
					local iTurnsOnSite:number = pGovernor:GetTurnsOnSite();
					local iTurnsToEstablish:number = pGovernor:GetTurnsToEstablish();
					local iTurnsUntilEstablished:number = iTurnsToEstablish - iTurnsOnSite;
					establishInfo = Locale.Lookup("LOC_GOVERNORS_SCREEN_GOVERNOR_TURNS_UNTIL_ESTABLISHED", iTurnsUntilEstablished);
				end
			end

			local playerConfig = PlayerConfigurations[playerID];
			if establishInfo then
				otherGovernorsTT = otherGovernorsTT .. Locale.Lookup("LOC_CITY_BANNER_HAS_AMBASSADOR_TOOLTIP_ENTRY_FULL", Locale.Lookup(playerConfig:GetCivilizationDescription()), Locale.Lookup(playerConfig:GetPlayerName()), establishInfo);
			else
				otherGovernorsTT = otherGovernorsTT .. Locale.Lookup("LOC_CITY_BANNER_HAS_AMBASSADOR_TOOLTIP_ENTRY_LIMITED", Locale.Lookup(playerConfig:GetCivilizationDescription()), Locale.Lookup(playerConfig:GetPlayerName()));
			end
		end
	end

	-- We must be a city state, multiple diplomats can be slotted here (but only show one icon)
	if otherGovernors > 0 then
		local instance:table = self.m_StatGovernorIM:GetInstance();
		instance.FillMeter:SetPercent(1.0);
		instance.FillMeter:SetTexture(IconManager:FindIconAtlas("ICON_GOVERNOR_OTHER_AMBASSADORS_FILL", 32));
		instance.SlotMeter:SetPercent(0);
		instance.UnknownGovernor:SetHide(true);
		instance.FillMeter:SetToolTipString(otherGovernorsTT);
		instance.TurnsLeft:SetText("");
		instance.NumOfAmbassadors:SetText(otherGovernors > 1 and tostring(otherGovernors) or "");

		instance.Button:RegisterCallback(Mouse.eLClick, OnGovernorIconClicked);
		instance.Button:SetVoid1(pCity:GetOwner());
		instance.Button:SetVoid2(pCity:GetID());
	end
end

-- ===========================================================================
function CityBanner:UpdateStats()
	local pDistrict:table = self:GetDistrict();
	local localPlayerID:number = Game.GetLocalPlayer();

	if (pDistrict ~= nil) then
		
		local districtHitpoints		:number = pDistrict:GetMaxDamage(DefenseTypes.DISTRICT_GARRISON);
		local currentDistrictDamage :number = pDistrict:GetDamage(DefenseTypes.DISTRICT_GARRISON);
		local wallHitpoints			:number = pDistrict:GetMaxDamage(DefenseTypes.DISTRICT_OUTER);
		local currentWallDamage		:number = pDistrict:GetDamage(DefenseTypes.DISTRICT_OUTER);
		local garrisonDefense		:number = math.floor(pDistrict:GetDefenseStrength() + 0.5);

		if self.m_Type == BANNERTYPE_CITY_CENTER then

			local pCity				:table = self:GetCity();
			local iCityOwner		:number = pCity:GetOwner();
			local pCityGrowth		:table  = pCity:GetGrowth();
			local pBuildQueue		:table  = pCity:GetBuildQueue();
			local populationIM		:table;

			if (localPlayerID == iCityOwner) then

				-- CITY POPULATION
				self:UpdatePopulation(true, pCity, pCityGrowth);

				-- CITY GOVERNOR
				self:UpdateGovernor(pCity);

				-- CITY PRODUCTION
				self.m_StatProductionIM:ResetInstances();

				local productionInstance:table = self.m_StatProductionIM:GetInstance();
				productionInstance.Button:RegisterCallback( Mouse.eLClick, OnProductionClick );
				productionInstance.Button:SetVoid1(iCityOwner);
				productionInstance.Button:SetVoid2(pCity:GetID());

				if (pBuildQueue ~= nil) then
					local productionpct			:number = 0;
					local currentProduction		:string;
					local currentProductionHash :number = pBuildQueue:GetCurrentProductionTypeHash();
					local prodTurnsLeft			:number;
					local progress				:number;
					local prodTypeName			:string;
					local pBuildingDef			:table;
					local pDistrictDef			:table;
					local pUnitDef				:table;
					local pProjectDef			:table;

					-- Attempt to obtain a hash for each item
					if currentProductionHash ~= 0 then
						pBuildingDef = GameInfo.Buildings[currentProductionHash];
						pDistrictDef = GameInfo.Districts[currentProductionHash];
						pUnitDef	 = GameInfo.Units[currentProductionHash];
						pProjectDef	 = GameInfo.Projects[currentProductionHash];
					end

					if( pBuildingDef ~= nil ) then
						currentProduction = pBuildingDef.Name;
						prodTypeName = pBuildingDef.BuildingType;
						prodTurnsLeft = pBuildQueue:GetTurnsLeft(pBuildingDef.BuildingType);
						progress = pBuildQueue:GetBuildingProgress(pBuildingDef.Index);
						productionpct = progress / pBuildQueue:GetBuildingCost(pBuildingDef.Index);
					elseif ( pDistrictDef ~= nil ) then
						currentProduction = pDistrictDef.Name;
						prodTypeName = pDistrictDef.DistrictType;
						prodTurnsLeft = pBuildQueue:GetTurnsLeft(pDistrictDef.DistrictType);
						progress = pBuildQueue:GetDistrictProgress(pDistrictDef.Index);
						productionpct = progress / pBuildQueue:GetDistrictCost(pDistrictDef.Index);
					elseif ( pUnitDef ~= nil ) then
						local eMilitaryFormationType = pBuildQueue:GetCurrentProductionTypeModifier();
						currentProduction = pUnitDef.Name;
						prodTypeName = pUnitDef.UnitType;
						prodTurnsLeft = pBuildQueue:GetTurnsLeft(pUnitDef.UnitType, eMilitaryFormationType);
						progress = pBuildQueue:GetUnitProgress(pUnitDef.Index);

						if (eMilitaryFormationType == MilitaryFormationTypes.STANDARD_FORMATION) then
							productionpct = progress / pBuildQueue:GetUnitCost(pUnitDef.Index);	
						elseif (eMilitaryFormationType == MilitaryFormationTypes.CORPS_FORMATION) then
							productionpct = progress / pBuildQueue:GetUnitCorpsCost(pUnitDef.Index);
							if (pUnitDef.Domain == "DOMAIN_SEA") then
								-- Concatenanting two fragments is not loc friendly.  This needs to change.
								currentProduction = Locale.Lookup(currentProduction) .. " " .. Locale.Lookup("LOC_UNITFLAG_FLEET_SUFFIX");
							else
								-- Concatenanting two fragments is not loc friendly.  This needs to change.
								currentProduction = Locale.Lookup(currentProduction) .. " " .. Locale.Lookup("LOC_UNITFLAG_CORPS_SUFFIX");
							end
						elseif (eMilitaryFormationType == MilitaryFormationTypes.ARMY_FORMATION) then
							productionpct = progress / pBuildQueue:GetUnitArmyCost(pUnitDef.Index);
							if (pUnitDef.Domain == "DOMAIN_SEA") then
								-- Concatenanting two fragments is not loc friendly.  This needs to change.
								currentProduction = Locale.Lookup(currentProduction) .. " " .. Locale.Lookup("LOC_UNITFLAG_ARMADA_SUFFIX");
							else
								-- Concatenanting two fragments is not loc friendly.  This needs to change.
								currentProduction = Locale.Lookup(currentProduction) .. " " .. Locale.Lookup("LOC_UNITFLAG_ARMY_SUFFIX");
							end
						end

						progress = pBuildQueue:GetUnitProgress(pUnitDef.Index);
						productionpct = progress / pBuildQueue:GetUnitCost(pUnitDef.Index);
					elseif (pProjectDef ~= nil) then
						currentProduction = pProjectDef.Name;
						prodTypeName = pProjectDef.ProjectType;
						prodTurnsLeft = pBuildQueue:GetTurnsLeft(pProjectDef.ProjectType);
						progress = pBuildQueue:GetProjectProgress(pProjectDef.Index);
						productionpct = progress / pBuildQueue:GetProjectCost(pProjectDef.Index);
					end

					if (currentProduction ~= nil) then
						productionpct = math.clamp(productionpct, 0, 1);

						productionInstance.FillMeter:SetHide(false);
						productionInstance.FillMeter:SetPercent(productionpct);

						productionInstance.SlotMeter:SetHide(false);
						productionInstance.SlotMeter:SetPercent(1 - productionpct);

						local productionTip				:string = Locale.Lookup("LOC_CITY_BANNER_PRODUCING", currentProduction);
						local productionTurnsLeftString :string;
						if prodTurnsLeft <= 0 then
							productionInstance.TurnsLeft:SetText("-");
							productionTurnsLeftString = "  " .. Locale.Lookup("LOC_CITY_BANNER_TURNS_LEFT_UNTIL_COMPLETE", "-");
						else
							productionTurnsLeftString = "  " .. Locale.Lookup("LOC_CITY_BANNER_TURNS_LEFT_UNTIL_COMPLETE", prodTurnsLeft);
							productionInstance.TurnsLeft:SetText(prodTurnsLeft);
						end
						productionTip = productionTip .. "[NEWLINE]" .. productionTurnsLeftString;
						productionInstance.Button:SetToolTipString(productionTip);
						productionInstance.Button:SetColor(0x00FFFFFF);
						
						if(prodTypeName ~= nil) then
							productionInstance.Icon:SetHide(false);
							productionInstance.Icon:SetIcon("ICON_"..prodTypeName);
						else
							UI.DataError("City has current production, but no prodTypeName");
						end
					else
						productionInstance.Button:SetColor(0xFFFFFFFF);
						productionInstance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_NO_PRODUCTION"));
						productionInstance.FillMeter:SetHide(true);
						productionInstance.SlotMeter:SetHide(true);
						productionInstance.Icon:SetHide(true);
					end
				end
			else
				self:UpdatePopulation(false, pCity, pCityGrowth);
				self:UpdateGovernor(pCity);
			end

			--- DEFENSE INFO ---
			local garrisonDefString :string = Locale.Lookup("LOC_CITY_BANNER_GARRISON_DEFENSE_STRENGTH");
			local defValue = garrisonDefense;
			local defTooltip = garrisonDefString .. ": " .. garrisonDefense;
			local healthTooltip :string = Locale.Lookup("LOC_CITY_BANNER_GARRISON_HITPOINTS", ((districtHitpoints-currentDistrictDamage) .. "/" .. districtHitpoints));
			if (wallHitpoints > 0) then
				self.m_Instance.DefenseIcon:SetHide(true);
				self.m_Instance.ShieldsIcon:SetHide(false);
				self.m_Instance.CityDefenseBarBacking:SetHide(false);
				self.m_Instance.CityHealthBarBacking:SetHide(false);
				self.m_Instance.CityDefenseBar:SetHide(false);
				healthTooltip = healthTooltip .. "[NEWLINE]" .. Locale.Lookup("LOC_CITY_BANNER_OUTER_DEFENSE_HITPOINTS", ((wallHitpoints-currentWallDamage) .. "/" .. wallHitpoints));
				self.m_Instance.CityDefenseBar:SetPercent((wallHitpoints-currentWallDamage) / wallHitpoints);
				self.m_Instance.CityDefenseBarBacking:SetToolTipString(healthTooltip);
			else
				self.m_Instance.CityDefenseBar:SetHide(true);
				self.m_Instance.CityDefenseBarBacking:SetHide(true);
				self.m_Instance.CityHealthBarBacking:SetHide(true);
			end
			self.m_Instance.DefenseNumber:SetText(defValue);
			self.m_Instance.DefenseNumber:SetToolTipString(defTooltip);
			self.m_Instance.CityHealthBarBacking:SetToolTipString(healthTooltip);
			self.m_Instance.CityHealthBarBacking:SetHide(false);
			if(districtHitpoints > 0) then
				self.m_Instance.CityHealthBar:SetPercent((districtHitpoints-currentDistrictDamage) / districtHitpoints);	
			else
				self.m_Instance.CityHealthBar:SetPercent(0);	
			end
			self:SetHealthBarColor();	
			
			if (((districtHitpoints-currentDistrictDamage) / districtHitpoints) == 1 and wallHitpoints == 0) then
				self.m_Instance.CityHealthBar:SetHide(true);
				self.m_Instance.CityHealthBarBacking:SetHide(true);
			else
				self.m_Instance.CityHealthBar:SetHide(false);
				self.m_Instance.CityHealthBarBacking:SetHide(false);
			end

			self:UpdateDetails();
			--------------------------------------
		else -- it should be a miniBanner
			
			if (self.m_Type == BANNERTYPE_ENCAMPMENT) then 
				self:UpdateEncampmentBanner();
			elseif (self.m_Type == BANNERTYPE_AERODROME) then
				self:UpdateAerodromeBanner();
			end
			
		end

	else  --it's a banner not associated with a district
		if (self.m_IsImprovementBanner) then
			local bannerPlot = Map.GetPlot(self.m_PlotX, self.m_PlotY);
			if (bannerPlot ~= nil) then
				if (self.m_Type == BANNERTYPE_AERODROME) then
					self:UpdateAerodromeBanner();
				elseif (self.m_Type == BANNERTYPE_MISSILE_SILO) then
					self:UpdateWMDBanner();
				end
			end
		end
	end
end

-- ===========================================================================
function SetDetailIcon(instance:table, icon:string, tooltip:string)
	instance.Icon:SetHide(icon == nil);
	if icon then instance.Icon:SetIcon(icon); end
	instance.Icon:SetToolTipString(tooltip and Locale.Lookup(tooltip) or "");
end

-- ===========================================================================
function CityBanner:UpdateDetails()
	
	local pCity:table = self:GetCity();
	local pDistrict:table = self:GetDistrict();
	if pCity and pDistrict then
		local cityOwner:number = pCity:GetOwner();
		local localPlayerID:number = Game.GetLocalPlayer();

		-- RESET INSTANCES
		self.m_DetailStatusIM:ResetInstances();
		self.m_DetailEffectsIM:ResetInstances();
		
		local bHasQuests: boolean = false;
		local questsManager: table = Game.GetQuestsManager();
		local questTooltip: string = Locale.Lookup("LOC_CITY_STATES_QUESTS");
		if questsManager then
			for questInfo in GameInfo.Quests() do
				if questsManager:HasActiveQuestFromPlayer(localPlayerID, cityOwner, questInfo.Index) then
					bHasQuests = true;
					questTooltip = questTooltip .. "[NEWLINE]" .. questInfo.IconString .. questsManager:GetActiveQuestName(localPlayerID, cityOwner, questInfo.Index);
				end
			end
		end

		if bHasQuests then
			SetDetailIcon(self.m_DetailStatusIM:GetInstance(), "ICON_CITY_STATUS_QUEST", questTooltip);
		end

		if m_isTradeSelectionActive then
			local pCityTrade:table = pCity:GetTrade();
			if pCityTrade:HasActiveTradingPost(localPlayer) then
				SetDetailIcon(self.m_DetailStatusIM:GetInstance(), "ICON_CITY_STATUS_TRADE_ACTIVE", "LOC_CITY_BANNER_ACTIVE_TRADING");
			elseif pCityTrade:HasInactiveTradingPost(localPlayer) then
				SetDetailIcon(self.m_DetailStatusIM:GetInstance(), "ICON_CITY_STATUS_TRADE_INACTIVE", "LOC_CITY_BANNER_INACTIVE_TRADING");
			end
		end

		if pDistrict:CanAttack() then
			SetDetailIcon(self.m_DetailStatusIM:GetInstance(), "ICON_CITY_STATUS_STRIKE", "LOC_CITY_BANNER_CAN_STRIKE");
		end
		
		self.m_Instance.CityDetailsStatus:CalculateSize();

		-- Update under siege icon
		if pDistrict:IsUnderSiege() then
			SetDetailIcon(self.m_DetailEffectsIM:GetInstance(), "ICON_CITY_EFFECTS_SIEGE", "LOC_HUD_REPORTS_STATUS_UNDER_SEIGE");
		end

		if cityOwner == localPlayerID then
			local pCityGrowth:table = pCity:GetGrowth();

			-- Update occupied icon
			if pCity:IsOccupied() then
				SetDetailIcon(self.m_DetailEffectsIM:GetInstance(), "ICON_CITY_EFFECTS_OCCUPIED", "LOC_HUD_CITY_GROWTH_OCCUPIED");
			end

			-- Update insufficient housing icon
			if pCityGrowth:GetHousing() < pCity:GetPopulation() then
				SetDetailIcon(self.m_DetailEffectsIM:GetInstance(), "ICON_CITY_EFFECTS_HOUSING", "LOC_CITY_BANNER_HOUSING_INSUFFICIENT");
			end

			-- Update insufficient amenities icon
			if pCityGrowth:GetAmenitiesNeeded() > pCityGrowth:GetAmenities() then
				SetDetailIcon(self.m_DetailEffectsIM:GetInstance(), "ICON_CITY_EFFECTS_AMENITIES", "LOC_CITY_BANNER_AMENITIES_INSUFFICIENT");
			end	
		end

		self.m_Instance.CityDetailsEffects:CalculateSize();
	end

	self:Resize();
end

-- ===========================================================================
--	Round to X decimal places -- do we have a function for this already?
-- ===========================================================================
function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult;
end

function OnCityBannerLookAt( playerID:number, cityID:number )
	local pPlayer = Players[playerID];
	if (pPlayer == nil) then
		return;
	end
	
	local pCity = pPlayer:GetCities():FindID(cityID);
	if (pCity == nil) then
		return;
	end

	UI.LookAtPlotScreenPosition( pCity:GetX(), pCity:GetY(), 0.5, 0.5 );
end

-- ===========================================================================
function OnCityBannerClick( playerID:number, cityID:number )
	local pPlayer = Players[playerID];
	if (pPlayer == nil) then
		return;
	end
	
	local pCity = pPlayer:GetCities():FindID(cityID);
	if (pCity == nil) then
		return;
	end

	if (pPlayer:IsFreeCities()) then
		UI.LookAtPlotScreenPosition( pCity:GetX(), pCity:GetY(), 0.5, 0.5 );
		return;
	end
	
	local localPlayerID;
	if (WorldBuilder.IsActive()) then
		localPlayerID = playerID;	-- If WorldBuilder is active, allow the user to select the city
	else
		localPlayerID = Game.GetLocalPlayer();
	end

	if (pPlayer:GetID() == localPlayerID) then
		UI.SelectCity( pCity );
	elseif(localPlayerID == PlayerTypes.OBSERVER 
			or localPlayerID == PlayerTypes.NONE 
			or pPlayer:GetDiplomacy():HasMet(localPlayerID)) then
		
		local pPlayerConfig :table		= PlayerConfigurations[playerID];
		local isMinorCiv	:boolean	= pPlayerConfig:GetCivilizationLevelTypeID() ~= CivilizationLevelTypes.CIVILIZATION_LEVEL_FULL_CIV;
		--print("clicked player " .. playerID .. " city.  IsMinor?: ",isMinorCiv);

		if UI.GetInterfaceMode() == InterfaceModeTypes.MAKE_TRADE_ROUTE then
			local plotID = Map.GetPlotIndex(pCity:GetX(), pCity:GetY());
			LuaEvents.CityBannerManager_MakeTradeRouteDestination( plotID );	
		else		
			if isMinorCiv then
				if UI.GetInterfaceMode() ~= InterfaceModeTypes.SELECTION then
					UI.SetInterfaceMode(InterfaceModeTypes.SELECTION);
				end
				LuaEvents.CityBannerManager_RaiseMinorCivPanel( playerID );	-- Go directly to a city-state
			else
				LuaEvents.CityBannerManager_TalkToLeader( playerID );
			end
		end
		
	end
end

-- ===========================================================================
function OnMiniBannerClick( playerID, districtID )
	local pPlayer = Players[playerID];
	if (pPlayer == nil) then
		return;
	end

	local pDistrict = pPlayer:GetDistricts():FindID(districtID);
	if (pDistrict == nil) then
		return;
	end

	if (pPlayer:GetID() == Game.GetLocalPlayer()) then
		UI.DeselectAll();
		UI.SelectDistrict( pDistrict );
		--handle air unit menu here
	end
end

-- ===========================================================================
function OnProductionClick( playerID, cityID )
	local pPlayer = Players[playerID];
	if (pPlayer == nil) then
		return;
	end
	
	local pCity = pPlayer:GetCities():FindID(cityID);
	if (pCity == nil) then
		return;
	end
	
	UI.SelectCity( pCity );											
	--UI.SelectCity( pCity, false );										-- Don't auto center
	--UI.LookAtPlotScreenPosition( pCity:GetX(), pCity:GetY(), 0.40, 0.5 );	-- Just a little bit to the right since production panel is opening


	LuaEvents.CityBannerManager_ProductionToggle();
end

-- ===========================================================================
function CityBanner:GetCity()
	local pCity : table = self.m_Player:GetCities():FindID(self.m_CityID);
	return pCity;
end

-- ===========================================================================
function CityBanner:GetDistrict()
	local pDistrict : table = self.m_Player:GetDistricts():FindID(self.m_DistrictID);
	return pDistrict;
end

-- ===========================================================================
function CityBanner:GetImprovementInfo()
	local tImprovementInfo = {
		LocX						= self.m_PlotX,
		LocY						= self.m_PlotY,
		ImprovementOwner			= -1,
		AirUnits					= {},
	};
	return tImprovementInfo;
end

-- ===========================================================================
function CityBanner:SetFogState( fogState : number )

	if( fogState == PLOT_HIDDEN ) then
		self:SetHide( true );
	else
		self:SetHide( false );

		--If this is an Aerodrome we need to hide the numbers and dropdown if in FOW
		if( self.m_Type == BANNERTYPE_AERODROME) then
			if( fogState == PLOT_REVEALED ) then
				self.m_Instance.AerodromeBase:SetHide(true);
				self.m_Instance.UnitListPopup:SetDisabled(true);
			else
				self.m_Instance.AerodromeBase:SetHide(false);
				self.m_Instance.UnitListPopup:SetDisabled(not self.m_UnitListEnabled);
			end
		end

	end
	self.m_FogState = fogState;
end

-- ===========================================================================
function CityBanner:SetHide( bHide : boolean )
	self.m_IsCurrentlyVisible = not bHide;
	self:UpdateVisibility();
end

-- ===========================================================================
function CityBanner:UpdateVisibility()

	local bVisible = self.m_IsCurrentlyVisible and not self.m_IsForceHide;
	self.m_Instance.Anchor:SetHide(not bVisible);
	self:UpdateLoyaltyWarning();
end

-- ===========================================================================
function CityBanner:UpdateName()
	if (self.m_Type == BANNERTYPE_CITY_CENTER) then
		local pCity : table = self:GetCity();
		if pCity ~= nil then
			local cityName:string = pCity:GetName();

			local tooltip:string = "";
			local owner:number = pCity:GetOwner();
			local pPlayer:table  = Players[owner];
			if pPlayer and pPlayer:IsMajor() then
				tooltip = Locale.Lookup("LOC_CITY_BANNER_TT", cityName, PlayerConfigurations[owner]:GetCivilizationShortDescription());
			else
				tooltip = Locale.Lookup(cityName);
			end
			
			self.m_Instance.CityName:SetText( Locale.ToUpper(cityName) );
			self.m_Instance.CityBannerButton:SetToolTipString( tooltip );
			self:UpdateInfo( pCity );
			self:Resize();
		end
	end
end
-- ===========================================================================
function OnCivIconClicked(playerID: number, cityID: number)
	if (playerID == Game.GetLocalPlayer()) then
		OnCityBannerClick(playerID, cityID);
		LuaEvents.CityPanel_ToggleOverviewLoyalty();
	else
		OnCityBannerLookAt(playerID, cityID);
		LuaEvents.OnViewLoyaltyLens();
	end
end

-- ===========================================================================
function OnReligionIconClicked(playerID: number, cityID: number)
	if (playerID == Game.GetLocalPlayer()) then
		OnCityBannerClick(playerID, cityID);
		LuaEvents.CityPanel_ToggleOverviewReligion();
	else
		OnCityBannerLookAt(playerID, cityID);
		LuaEvents.OnViewReligionLens();
	end
end

-- ===========================================================================
function CityBanner:UpdateInfo( pCity : table )
	
	self.m_CivIconInstance = nil;
	self.m_InfoIconIM:ResetInstances();
	self.m_InfoConditionIM:ResetInstances();

	if pCity ~= nil then
		local playerID		:number = pCity:GetOwner();
		local cityID		:number = pCity:GetID();
		local pPlayer		:table	= Players[playerID];
		local pPlayerConfig	:table	= PlayerConfigurations[playerID];

		-- CAPITAL ICON
		if pPlayer then
			local instance:table = self.m_InfoIconIM:GetInstance();
			instance.Button:RegisterCallback(Mouse.eLClick, OnCityBannerClick);
			instance.Button:SetVoid1(playerID);
			instance.Button:SetVoid2(cityID);

			if pPlayer:IsMajor() then
				if pCity:IsOriginalCapital() and pCity:GetOriginalOwner() == pCity:GetOwner() then
					instance.Icon:SetIcon("ICON_CITY_CAPITAL");
					instance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_ORIGINAL_CAPITAL_TT", pPlayerConfig:GetCivilizationShortDescription()));
				elseif pCity:IsCapital() then
					instance.Icon:SetIcon("ICON_NEW_CAPITAL");
					instance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_NEW_CAPITAL_TT", pPlayerConfig:GetCivilizationShortDescription()));
				else
					instance.Icon:SetIcon("ICON_OTHER_CITIES");
					instance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_OTHER_CITY_TT", pPlayerConfig:GetCivilizationShortDescription()));
				end
			elseif pPlayer:IsFreeCities() then
				instance.Icon:SetIcon("ICON_CIVILIZATION_FREE_CITIES");
				instance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_FREE_CITY_TT"));
			else
				instance.Icon:SetIcon("ICON_CITY_STATE");
				instance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_CITY_STATE_TT"));
			end

			-- ORIGINAL OWNER CAPITAL ICON
			if pCity:GetOwner() ~= pCity:GetOriginalOwner() and pCity:IsOriginalCapital() then
				local pOriginalOwner:table = Players[pCity:GetOriginalOwner()];
				-- Only show the captured capital icon for major civs
				if pOriginalOwner:IsMajor() then
					local instance:table = self.m_InfoIconIM:GetInstance();
					instance.Icon:SetIcon("ICON_CAPTURED_CAPITAL");
					local pOriginalOwnerConfig:table = PlayerConfigurations[pCity:GetOriginalOwner()];
					instance.Button:SetToolTipString(Locale.Lookup("LOC_CITY_BANNER_CAPTURED_CAPITAL_TT", pOriginalOwnerConfig:GetCivilizationShortDescription()));
					instance.Button:RegisterCallback(Mouse.eLClick, OnCityBannerClick);
					instance.Button:SetVoid1(pCity:GetOriginalOwner());
					instance.Button:SetVoid2(cityID);
				end
			end
		end

		-- CIV ICON
		local leader:string = pPlayerConfig:GetLeaderTypeName();
		if leader ~= nil then
			if GameInfo.CivilizationLeaders[leader] == nil then
				UI.DataError("Banners found a leader \""..leader.."\" which is not/no longer in the game; icon may be whack.");
			else
				local civType:string = GameInfo.CivilizationLeaders[leader].CivilizationType;
				if civType ~= nil then
					self.m_CivIconInstance = self.m_InfoConditionIM:GetInstance();
					self.m_CivIconInstance.Icon:SetIcon("ICON_" .. civType);
	
					local tooltip:string = Locale.Lookup("LOC_CITY_BELONGS_TT", pPlayerConfig:GetCivilizationShortDescription());
					local isLoyaltyRising:boolean = false;
					local isLoyaltyFalling:boolean = false;
					local pCulturalIdentity:table = pCity:GetCulturalIdentity();
					if pCulturalIdentity then
						local conversionOutcome:number = pCulturalIdentity:GetConversionOutcome();
						local turnsToConversion:number = pCulturalIdentity:GetTurnsToConversion();
						if conversionOutcome == IdentityConversionOutcome.GAINING_LOYALTY then
							isLoyaltyRising = true;
							if (turnsToConversion > 0) then
								tooltip = tooltip .. "[NEWLINE]" .. Locale.Lookup("LOC_HUD_CITY_GAINING_LOYALTY", turnsToConversion);
							else
								tooltip = tooltip .. "[NEWLINE]" .. Locale.Lookup("LOC_HUD_CITY_FULL_LOYALTY");
							end
						elseif conversionOutcome == IdentityConversionOutcome.LOSING_LOYALTY then
							isLoyaltyFalling = true;
							tooltip = tooltip .. "[NEWLINE]" .. Locale.Lookup("LOC_HUD_CITY_LOSING_LOYALTY", turnsToConversion);
						else
							tooltip = tooltip .. "[NEWLINE]" .. Locale.Lookup("LOC_HUD_CITY_STABLE_LOYALTY");
						end
					end
					self.m_CivIconInstance.ConditionRising:SetHide(not isLoyaltyRising or isLoyaltyFalling);
					self.m_CivIconInstance.ConditionFalling:SetHide(not isLoyaltyFalling or isLoyaltyRising);
					self.m_CivIconInstance.Button:SetToolTipString(tooltip);
	
					self.m_CivIconInstance.Button:RegisterCallback( Mouse.eLClick, OnCivIconClicked );
					self.m_CivIconInstance.Button:SetVoid1(playerID);
					self.m_CivIconInstance.Button:SetVoid2(cityID);
				end
			end
		end

		-- RELIGION ICON
		local pCityReligion:table = pCity:GetReligion();
		local eMajorityReligion:number = self.m_eMajorityReligion;
		if (eMajorityReligion > 0) then
			local instance:table = self.m_InfoConditionIM:GetInstance();
			local majorityReligionColor:number = UI.GetColorValue(GameInfo.Religions[eMajorityReligion].Color);
			instance.Icon:SetColor(majorityReligionColor and majorityReligionColor or COLOR_HOLY_SITE);
			instance.Icon:SetIcon("ICON_" .. GameInfo.Religions[eMajorityReligion].ReligionType);
			instance.Button:SetToolTipString(Locale.Lookup("LOC_HUD_CITY_RELIGION_TT", Game.GetReligion():GetName(eMajorityReligion)));
			instance.Button:RegisterCallback( Mouse.eLClick, OnReligionIconClicked );
			instance.Button:SetVoid1(playerID);
			instance.Button:SetVoid2(cityID);

			-- Get a list of religions present in this city
			local otherReligionPressure:number = 0;
			local pReligionsInCity:table = pCityReligion:GetReligionsInCity();
			for _, cityReligion in pairs(pReligionsInCity) do
				local religion:number = cityReligion.Religion;
				if religion >= 0 and religion ~= eMajorityReligion then
					otherReligionPressure = otherReligionPressure + pCityReligion:GetTotalPressureOnCity(religion);
				end
			end

			local majorityPressure:number = pCityReligion:GetTotalPressureOnCity(eMajorityReligion);

			local isPressureRising:boolean = majorityPressure > otherReligionPressure;
			local isPressureFalling:boolean = majorityPressure < otherReligionPressure;

			instance.ConditionRising:SetHide(not isPressureRising or isPressureFalling);
			instance.ConditionFalling:SetHide(not isPressureFalling or isPressureRising);
		else
			local activePantheon:number = pCityReligion:GetActivePantheon();
			if (activePantheon >= 0) then
				local instance:table = self.m_InfoIconIM:GetInstance();
				instance.Icon:SetIcon("ICON_" .. GameInfo.Religions[0].ReligionType);
				instance.Icon:SetColor(COLOR_HOLY_SITE);
				instance.Button:SetToolTipString(Locale.Lookup("LOC_HUD_CITY_PANTHEON_TT", GameInfo.Beliefs[activePantheon].Name));
				instance.Button:RegisterCallback( Mouse.eLClick, OnReligionIconClicked );
				instance.Button:SetVoid1(playerID);
				instance.Button:SetVoid2(cityID);
			end
		end

		-- LOYALTY WARNING
		self:UpdateLoyaltyWarning();
	end

	self:Resize();
end

-- ===========================================================================
function CityBanner:UpdateReligion()

	local pCity				:table = self:GetCity();
	local pCityReligion		:table = pCity:GetReligion();
	local localPlayerID		:number = Game.GetLocalPlayer();
	local eMajorityReligion	:number = pCityReligion:GetMajorityReligion();

	self.m_eMajorityReligion = eMajorityReligion;
	self:UpdateInfo(pCity);

	local cityInst			:table = self.m_Instance;
	local religionInfo		:table = cityInst.ReligionInfo;
	local religionsInCity	:table = pCityReligion:GetReligionsInCity();

	-- Hide the meter and bail out if the religion lens isn't active
	if(not m_isReligionLensActive or table.count(religionsInCity) == 0) then
		if religionInfo then
			religionInfo.ReligionInfoContainer:SetHide(true);
		end
		return;
	end

	-- Update religion icon + religious pressure animation
	local majorityReligionColor:number = COLOR_RELIGION_DEFAULT;
	if(eMajorityReligion >= 0) then
		majorityReligionColor = UI.GetColorValue(GameInfo.Religions[eMajorityReligion].Color);
	end
	
	-- Preallocate total fill so we can stagger the meters
	local totalFillPercent:number = 0;
	local iCityPopulation:number = pCity:GetPopulation();

	-- Get a list of religions present in this city
	local activeReligions:table = {};
	local pReligionsInCity:table = pCityReligion:GetReligionsInCity();
	for _, cityReligion in pairs(pReligionsInCity) do
		local religion:number = cityReligion.Religion;
		if(religion >= 0) then
			local followers:number = cityReligion.Followers;
			local fillPercent:number = followers / iCityPopulation;
			totalFillPercent = totalFillPercent + fillPercent;

			table.insert(activeReligions, {
				Religion=religion,
				Followers=followers,
				Pressure=pCityReligion:GetTotalPressureOnCity(religion),
				LifetimePressure=cityReligion.Pressure,
				FillPercent=fillPercent,
				Color=GameInfo.Religions[religion].Color });
		end
	end
	
	-- Sort religions by largest number of followers
	table.sort(activeReligions, function(a,b) return a.Followers > b.Followers; end);

	-- After sort update accumulative fill percent
	local accumulativeFillPercent = 0.0;
	for i, religion in ipairs(activeReligions) do
		accumulativeFillPercent = accumulativeFillPercent + religion.FillPercent;
		religion.AccumulativeFillPercent = accumulativeFillPercent;
	end

	if(table.count(activeReligions) > 0) then
		local localPlayerVis:table = PlayersVisibility[localPlayerID];
		if (localPlayerVis ~= nil) then
			-- Holy sites get a different color and texture
			local holySitePlotIDs:table = {};
			local cityDistricts:table = pCity:GetDistricts();
			local playerDistricts:table = self.m_Player:GetDistricts();
			for i, district in cityDistricts:Members() do
				local districtType:string = GameInfo.Districts[district:GetType()].DistrictType;
				if(districtType == "DISTRICT_HOLY_SITE") then
					local locX:number = district:GetX();
					local locY:number = district:GetY();
					if localPlayerVis:IsVisible(locX, locY) then
						local plot:table  = Map.GetPlot(locX, locY);
						local holySiteFaithYield:number = district:GetReligionHealRate();
						SpawnHolySiteIconAtLocation(locX, locY, "+" .. holySiteFaithYield);
						holySitePlotIDs[plot:GetIndex()] = true;
					end
					break;
				end
			end

			-- Color hexes in this city the same color as religion
			local plots:table = Map.GetCityPlots():GetPurchasedPlots(pCity);
			if(table.count(plots) > 0) then
				UILens.SetLayerHexesColoredArea( LensLayers.HEX_COLORING_RELIGION, localPlayerID, plots, majorityReligionColor );
			end
		end
	end

	if religionInfo then
		-- Create or reset icon instance manager
		local iconIM:table = cityInst[DATA_FIELD_RELIGION_ICONS_IM];
		if(iconIM == nil) then
			iconIM = InstanceManager:new("ReligionIconInstance", "ReligionIconButtonBacking", religionInfo.ReligionInfoIconStack);
			cityInst[DATA_FIELD_RELIGION_ICONS_IM] = iconIM;
		else
			iconIM:ResetInstances();
		end

		-- Create or reset follower list instance manager
		local followerListIM:table = cityInst[DATA_FIELD_RELIGION_FOLLOWER_LIST_IM];
		if(followerListIM == nil) then
			followerListIM = InstanceManager:new("ReligionFollowerListInstance", "ReligionFollowerListContainer", religionInfo.ReligionFollowerListStack);
			cityInst[DATA_FIELD_RELIGION_FOLLOWER_LIST_IM] = followerListIM;
		else
			followerListIM:ResetInstances();
		end

		-- Create or reset pop chart instance manager
		local popChartIM:table = cityInst[DATA_FIELD_RELIGION_POP_CHART_IM];
		if(popChartIM == nil) then
			popChartIM = InstanceManager:new("ReligionPopChartInstance", "PopChartMeter", religionInfo.ReligionPopChartContainer);
			cityInst[DATA_FIELD_RELIGION_POP_CHART_IM] = popChartIM;
		else
			popChartIM:ResetInstances();
		end

		local populationChartTooltip:string = RELIGION_POP_CHART_TOOLTIP_HEADER;

		-- Add religion icons for each active religion
		for i,religionInfo in ipairs(activeReligions) do
			local religionDef:table = GameInfo.Religions[religionInfo.Religion];

			local icon = "ICON_" .. religionDef.ReligionType;
			local religionColor = UI.GetColorValue(religionDef.Color);
		
			-- The first index is the predominant religion. Label it as such.
			local religionName = "";
			if i == 1 then
				religionName = Locale.Lookup("LOC_CITY_BANNER_PREDOMINANT_RELIGION", Game.GetReligion():GetName(religionDef.Index));
			else
				religionName = Game.GetReligion():GetName(religionDef.Index);
			end

			-- Add icon to main icon list
			local iconInst:table = iconIM:GetInstance();
			iconInst.ReligionIconButton:SetIcon(icon);
			iconInst.ReligionIconButton:SetColor(religionColor);
			iconInst.ReligionIconButtonBacking:SetColor(religionColor);
			iconInst.ReligionIconButtonBacking:SetToolTipString(religionName);

			-- Add followers to detailed info list
			local followerListInst:table = followerListIM:GetInstance();
			followerListInst.ReligionFollowerIcon:SetIcon(icon);
			followerListInst.ReligionFollowerIcon:SetColor(religionColor);
			followerListInst.ReligionFollowerIconBacking:SetColor(religionColor);
			followerListInst.ReligionFollowerCount:SetText(religionInfo.Followers);
			followerListInst.ReligionFollowerPressure:SetText(Locale.Lookup("LOC_CITY_BANNER_RELIGIOUS_PRESSURE", Round(religionInfo.Pressure)));

			-- Add the follower tooltip to the population chart tooltip
			local followerTooltip:string = Locale.Lookup("LOC_CITY_BANNER_FOLLOWER_PRESSURE_TOOLTIP", religionName, religionInfo.Followers, Round(religionInfo.LifetimePressure));
			followerListInst.ReligionFollowerIconBacking:SetToolTipString(followerTooltip);
			populationChartTooltip = populationChartTooltip .. "[NEWLINE][NEWLINE]" .. followerTooltip;
		end

		religionInfo.ReligionPopChartContainer:SetToolTipString(populationChartTooltip);
	
		religionInfo.ReligionFollowerListStack:CalculateSize();
		religionInfo.ReligionFollowerListScrollPanel:CalculateInternalSize();
		religionInfo.ReligionFollowerListScrollPanel:ReprocessAnchoring();

		-- Add populations to pie chart in reverse order
		for i = #activeReligions, 1, -1 do
			local religionInfo = activeReligions[i];
			local religionColor = UI.GetColorValue(religionInfo.Color);

			local popChartInst:table = popChartIM:GetInstance();
			popChartInst.PopChartMeter:SetPercent(religionInfo.AccumulativeFillPercent);
			popChartInst.PopChartMeter:SetColor(religionColor);
		end

		-- Update population pie chart majority religion icon
		if (eMajorityReligion > 0) then
			local iconName : string = "ICON_" .. GameInfo.Religions[eMajorityReligion].ReligionType;
			religionInfo.ReligionPopChartIcon:SetIcon(iconName);
			religionInfo.ReligionPopChartIcon:SetHide(false);
		else
			religionInfo.ReligionPopChartIcon:SetHide(true);
		end

		-- Show what religion we will eventually turn into
		local nextReligion = pCityReligion:GetNextReligion();
		local turnsTillNextReligion:number = pCityReligion:GetTurnsToNextReligion();
		if nextReligion and nextReligion ~= -1 and turnsTillNextReligion > 0 then
			local pNextReligionDef:table = GameInfo.Religions[nextReligion];

			-- Religion icon
			if religionInfo.ConvertingReligionIcon then
				local religionIcon = "ICON_" .. pNextReligionDef.ReligionType;
				religionInfo.ConvertingReligionIcon:SetIcon(religionIcon);
				local religionColor = UI.GetColorValue(pNextReligionDef.Color);
				religionInfo.ConvertingReligionIcon:SetColor(religionColor);
				religionInfo.ConvertingReligionIconBacking:SetColor(religionColor);
				religionInfo.ConvertingReligionIconBacking:SetToolTipString(Locale.Lookup(pNextReligionDef.Name));
			end

			-- Converting text
			local convertString = Locale.Lookup("LOC_CITY_BANNER_CONVERTS_IN_X_TURNS", turnsTillNextReligion);
			religionInfo.ConvertingReligionLabel:SetText(convertString);
			religionInfo.ReligionConversionTurnsStack:SetHide(false);

			-- If the turns till conversion are less than 10 play the warning flash animation
			religionInfo.ConvertingSoonAlphaAnim:SetToBeginning();
			if turnsTillNextReligion <= 10 then
				religionInfo.ConvertingSoonAlphaAnim:Play();
			else
				religionInfo.ConvertingSoonAlphaAnim:Stop();
			end
		else
			religionInfo.ReligionConversionTurnsStack:SetHide(true);
		end

		-- Show how much religion this city is exerting outwards
		local outwardReligiousPressure = pCityReligion:GetPressureFromCity();
		religionInfo.ExertedReligiousPressure:SetText(Locale.Lookup("LOC_CITY_BANNER_RELIGIOUS_PRESSURE", Round(outwardReligiousPressure)));

		-- Reset buttons to default state
		religionInfo.ReligionInfoButton:SetHide(false);
		religionInfo.ReligionInfoDetailedButton:SetHide(true);

		-- Register callbacks to open/close detailed info
		religionInfo.ReligionInfoButton:RegisterCallback( Mouse.eLClick, function() OnReligionInfoButtonClicked(religionInfo, pCity); end);
		religionInfo.ReligionInfoDetailedButton:RegisterCallback( Mouse.eLClick, function() OnReligionInfoDetailedButtonClicked(religionInfo, pCity); end);

		religionInfo.ReligionInfoContainer:SetHide(false);
	end
end

-- ===========================================================================
function OnReligionInfoButtonClicked( religionInfoance:table, pCity:table )
	if (m_preligionInfoance ~= nil) then
		m_preligionInfoance.ReligionInfoButton:SetHide(false);
		m_preligionInfoance.ReligionInfoDetailedButton:SetHide(true);
	end

	religionInfoance.ReligionInfoButton:SetHide(true);
	religionInfoance.ReligionInfoDetailedButton:SetHide(false);
	UILens.FocusCity(LensLayers.HEX_COLORING_RELIGION, pCity);
	m_preligionInfoance = religionInfoance;
end

-- ===========================================================================
function OnReligionInfoDetailedButtonClicked( religionInfoance:table, pCity:table )
	UI.AssertMsg(m_preligionInfoance == religionInfoance, "more than one panel was open");
	religionInfoance.ReligionInfoButton:SetHide(false);
	religionInfoance.ReligionInfoDetailedButton:SetHide(true);
	UILens.UnFocusCity(LensLayers.HEX_COLORING_RELIGION, pCity);
	m_preligionInfoance = nil;
end

-- ===========================================================================
function SpawnHolySiteIconAtLocation( locX : number, locY:number, label:string )
	local iconInst:table = m_HolySiteIconsIM:GetInstance();

	local xOffset:number = -4;	--offset to center UI element on tile
	local yOffset:number = 4;	--offset to center UI element on tile
	local zOffset:number = 10;	--offset for 3D world view
	if (UI.GetWorldRenderView() == WorldRenderView.VIEW_2D) then
		zOffset = 0;
	end

	local worldX:number, worldY:number, worldZ:number = UI.GridToWorld( locX, locY );
	iconInst.Anchor:SetWorldPositionVal( worldX + xOffset, worldY + yOffset, worldZ + zOffset );
	iconInst.HolySiteLabel:SetText("[ICON_FaithLarge]"..label);
	iconInst.Anchor:SetSizeX(iconInst.HolySiteBacking:GetSizeX());

	iconInst.Anchor:SetToolTipString(Locale.Lookup("LOC_UI_RELIGION_HOLY_SITE_BONUS_TT", label));
end

-- ===========================================================================
function CityBanner:UpdateLoyalty()

	local instance:table = self.m_Instance.LoyaltyInfo;
	if instance then
		if not m_isLoyaltyLensActive then
			instance.Top:SetHide(true);
			return;
		end
		local pCity:table = self:GetCity();
		if pCity then
			local pCityCulturalIdentity:table = pCity:GetCulturalIdentity();
			if pCityCulturalIdentity then
				local ownerID:number = pCity:GetOwner();

				local playerIdentitiesInCity = pCityCulturalIdentity:GetPlayerIdentitiesInCity();
				local cityIdentityPressures = pCityCulturalIdentity:GetCityIdentityPressures();
				local identitySourcesBreakdown = pCityCulturalIdentity:GetIdentitySourcesBreakdown();

				-- Update owner icon
				local pOwnerConfig:table = PlayerConfigurations[ownerID];
				local ownerIcon:string = "ICON_" .. pOwnerConfig:GetCivilizationTypeName();
				local ownerSecondaryColor, ownerPrimaryColor = UI.GetPlayerColors( ownerID );
				instance.OwnerCivIcon:SetIcon(ownerIcon);
				instance.OwnerCivIcon:SetColor(ownerPrimaryColor);
				instance.OwnerCivIcon:SetToolTipString(Locale.Lookup(pOwnerConfig:GetCivilizationDescription()));
				instance.OwnerCivIconBacking:SetColor(ownerSecondaryColor);
				instance.OwnerCivIconExtended:SetIcon(ownerIcon);
				instance.OwnerCivIconExtended:SetColor(ownerPrimaryColor);
				instance.OwnerCivIconBackingExtended:SetColor(ownerSecondaryColor);

				-- Update potential transfer player icon
				local transferPlayerID:number = pCityCulturalIdentity:GetPotentialTransferPlayer();
				if transferPlayerID ~= -1 then
					instance.TopCivIconBacking:SetHide(false);
					instance.TopCivIconBackingExtended:SetHide(false);

					local pTopConfig:table = PlayerConfigurations[transferPlayerID];
					local topIcon:string = "ICON_" .. pTopConfig:GetCivilizationTypeName();
					local topSecondaryColor, topPrimaryColor = UI.GetPlayerColors( transferPlayerID );
					instance.TopCivIcon:SetIcon(topIcon);
					instance.TopCivIcon:SetColor(topPrimaryColor);
					instance.TopCivIcon:SetToolTipString(Locale.Lookup(pTopConfig:GetCivilizationDescription()));

					instance.TopCivIconBacking:SetColor(topSecondaryColor);
					instance.TopCivIconExtended:SetIcon(topIcon);
					instance.TopCivIconExtended:SetColor(topPrimaryColor);
					instance.TopCivIconBackingExtended:SetColor(topSecondaryColor);
				else
					instance.TopCivIconBacking:SetHide(true);
					instance.TopCivIconBackingExtended:SetHide(false);
				end

				-- Update loyalty precentage
				local currentLoyalty:number = pCityCulturalIdentity:GetLoyalty();
				local maxLoyalty:number = pCityCulturalIdentity:GetMaxLoyalty();
				local loyalPercent:number = currentLoyalty / maxLoyalty;
				instance.LoyaltyFill:SetPercent(loyalPercent);
				instance.LoyaltyFillExtended:SetPercent(loyalPercent);
                
                
                
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------
                    -- My Mod UI
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------
                
                
                
                SetReligionModPerHeretics();
                local Playerid = pCity:GetComponentID().player
                local warexhaustionlevel = 0
                local ReligionMOdValue = 0
                local tooltipReligionMod2 = ""
                local TurnUntilRevoltAgain = 0
                local GovernmentMod2 = 0
                local GovermentTooltip = ""
                local ID = pCity:GetID()
                if Playerid ~= nil then
                    warexhaustionlevel = Warexhaustion[Playerid] 
                    local kOwnerCityId = Playerid.."."..pCity:GetID()
                    ReligionMOdValue = ReligionMod[kOwnerCityId]
                    tooltipReligionMod2 = tooltipReligionMod[kOwnerCityId]
                    local LastRevoltTurn =  CityLastRevolt[kOwnerCityId]
                    GovernmentMod2 = GovernmentMod[Playerid] 
                    GovermentTooltip = SetGovernmentModTooltip(Playerid)
                    if type(ReligionMOdValue) ~= "number" then
                        print ("ReligionMOdValue is "..type(ReligionMOdValue))
                        print (" player is "..Playerid.." CityID is "..ID)
                        ReligionMOdValue = 0
                        ReligionMod[kOwnerCityId] = 0
                    end
                    if type(warexhaustionlevel) ~= "number" then
                        print ("warexhaustionlevel is "..type(warexhaustionlevel))
                        warexhaustionlevel = 0
                        Warexhaustion[Playerid] = 0
                    end
                    if type(tooltipReligionMod2) ~= "string" then
                        print ("tooltipReligionMod2 is "..type(tooltipReligionMod2))
                        print (" player is "..Playerid.." CityID is "..ID)
                        tooltipReligionMod2 = "No Information"
                        tooltipReligionMod[kOwnerCityId] = ""
                    end
                    if  type(LastRevoltTurn) ~= "number" then
                         print ("LastRevoltTurn is "..type(LastRevoltTurn))
                        print (" player is "..Playerid.." CityID is "..ID)
                        LastRevoltTurn = -30
                        CityLastRevolt[kOwnerCityId] = -30    
                    end
                    TurnUntilRevoltAgain = LastRevoltTurn +30 - Currentturn () 
                    if type(GovernmentMod2) ~= "number" then
                         print ("GovernmentMod2 is "..type(GovernmentMod2))
                        print (" player is "..Playerid.." CityID is "..ID)
                        GovernmentMod2 = 0
                        GovernmentMod[Playerid] = 0
                    end
                    
                    
                end    
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------
                    -- My Mod UI end
                    ---------------------------------------------------------------------------------------------------------------------------------------------------------- 

  


				-- Determine which pressure font icon to use
				local loyaltyPerTurn:number = pCityCulturalIdentity:GetLoyaltyPerTurn() + warexhaustionlevel + ReligionMOdValue + GovernmentMod2;
				local loyaltyFontIcon:string = loyaltyPerTurn >= 0 and "[ICON_PressureUp]" or "[ICON_PressureDown]";

				-- Update loyalty percentage string

				local loyaltyText:string = Locale.Lookup("LOC_CULTURAL_IDENTITY_LOYALTY_PERCENTAGE", Round(currentLoyalty, 1), maxLoyalty, loyaltyFontIcon, Round(loyaltyPerTurn, 1));
				instance.LoyaltyPercentageLabel:SetText(loyaltyText);
				instance.LoyaltyPressureIcon:SetText(loyaltyFontIcon);
                if TurnUntilRevoltAgain > 0 then
                    local SetMax = Locale.Lookup("", 1000, maxLoyalty, loyaltyFontIcon, Round(loyaltyPerTurn, 1),"aaa");
                    
                    local LastrevoltTurnText = TurnUntilRevoltAgain.."[ICON_Turn] Turn untill this city can revolt again"
                   
                    local tooltip = "This City can't revolt cause of a Recent Revolt"
                    instance.LoyaltyPercentageLabel:SetText(SetMax);  
                    instance.LoyaltyFill:SetPercent(1);
				    instance.LoyaltyFillExtended:SetPercent(1);
                    instance.LastRevoltTurn:SetText(LastrevoltTurnText); 
                    instance.LastRevoltTurnTop:SetToolTipString(tooltip);
                    instance.LastRevoltTurnTop:SetHide(false)
                    else
                    instance.LastRevoltTurnTop:SetHide(true)
                end
				--Update Loyalty breakdown
				if self.m_LoyaltyBreakdownIM ~= nil then
					self.m_LoyaltyBreakdownIM:ResetInstances();

					--Populate the breakdown
					local localPlayerID = Game.GetLocalPlayer();
					local pCulturalIdentity = pCity:GetCulturalIdentity();
					local identitiesInCity = pCulturalIdentity:GetPlayerIdentitiesInCity();
					local firstIdentityInCity = next(identitiesInCity);
					if firstIdentityInCity == nil then
						--We have no presences, or we are the only one
						self.m_Instance.LoyaltyInfo.IdentityBreakdownStack:SetHide(true);
					else
						self.m_Instance.LoyaltyInfo.IdentityBreakdownStack:SetHide(false);
						table.sort(identitiesInCity, function(left, right)
							return left.IdentityTotal > right.IdentityTotal;
						end);

						local numInfluencers = 0;
						for i, playerPresence in ipairs(identitiesInCity) do
							if playerPresence.IdentityTotal ~= nil and playerPresence.IdentityTotal > 0 then
								if numInfluencers < 2 or playerPresence.Player == localPlayerID then
									numInfluencers = numInfluencers + 1;
									local instance = self.m_LoyaltyBreakdownIM:GetInstance();
									local localPlayer = Players[localPlayerID];
									local pPlayerConfig = PlayerConfigurations[playerPresence.Player];
									local civName = Locale.Lookup(pPlayerConfig:GetCivilizationShortDescription());
									local lineVal = (i == 1 and "[ICON_Bolt] " or "") .. Round(playerPresence.IdentityTotal, 1);
									local hasBeenMet = localPlayer:GetDiplomacy():HasMet(playerPresence.Player) or localPlayerID == playerPresence.Player;
									instance.LineTitle:SetText(hasBeenMet and civName or Locale.Lookup("LOC_LOYALTY_PANEL_UNMET_CIV"));
									instance.LineValue:SetText(lineVal);

									local civIconManager = CivilizationIcon:AttachInstance(instance.CivilizationIcon);
									civIconManager:UpdateIconFromPlayerID(playerPresence.Player);
								end
							end
						end
					end
					self.m_Instance.LoyaltyInfo.MainStack:CalculateSize();
				end
				

				-- Update loyalty pressure breakdown
				instance.FreeCityTop:SetHide(true);
				instance.CityStateTop:SetHide(true);
				for i , pressure in ipairs(identitySourcesBreakdown) do
					if pressure[PRESSURE_BREAKDOWN_TYPE_POPULATION_PRESSURE] then
						SetPressureBreakdownColumn(instance.PopulationPressureValue, instance.PopulationPressureFontIcon, Round(pressure[PRESSURE_BREAKDOWN_TYPE_POPULATION_PRESSURE], 1));
						local tooltip:string = Locale.Lookup("LOC_CULTURAL_IDENTITY_POPULATION_PRESSURE_TOOLTIP");
						local ownerPlayer = Players[ownerID];
						if (ownerPlayer ~= nil) then
							if (ownerPlayer:IsMajor()) then
								tooltip = tooltip .. "[NEWLINE][NEWLINE]" .. Locale.Lookup("LOC_CULTURAL_IDENTITY_POPULATION_PRESSURE_TOOLTIP_MAJOR_CIVS");
							else
								tooltip = tooltip .. "[NEWLINE][NEWLINE]" .. Locale.Lookup("LOC_CULTURAL_IDENTITY_POPULATION_PRESSURE_TOOLTIP_MINOR_CIVS");
							end
						end
						instance.PopulationTop:SetToolTipString(tooltip);
					elseif pressure[PRESSURE_BREAKDOWN_TYPE_GOVERNORS] then
						SetPressureBreakdownColumn(instance.GovernorPressureValue, instance.GovernorPressureFontIcon, Round(pressure[PRESSURE_BREAKDOWN_TYPE_GOVERNORS], 1));
					elseif pressure[PRESSURE_BREAKDOWN_TYPE_HAPPINESS] then
						SetPressureBreakdownColumn(instance.HappinessPressureValue, instance.HappinessPressureFontIcon, Round(pressure[PRESSURE_BREAKDOWN_TYPE_HAPPINESS], 1));
					elseif pressure[PRESSURE_BREAKDOWN_TYPE_OTHER] then
						SetPressureBreakdownColumn(instance.OtherPressureValue, instance.OtherPressureFontIcon, Round(pressure[PRESSURE_BREAKDOWN_TYPE_OTHER], 1));
					elseif pressure[PRESSURE_BREAKDOWN_TYPE_CITY_STATE_BONUS] then
						SetPressureBreakdownColumn(instance.CityStatePressureValue, instance.CityStatePressureFontIcon, Round(pressure[PRESSURE_BREAKDOWN_TYPE_CITY_STATE_BONUS], 1));
						instance.CityStateTop:SetHide(false);
					elseif pressure[PRESSURE_BREAKDOWN_TYPE_FREE_CITY_BONUS] then
						SetPressureBreakdownColumn(instance.FreeCityPressureValue, instance.FreeCityPressureFontIcon, Round(pressure[PRESSURE_BREAKDOWN_TYPE_FREE_CITY_BONUS], 1));
						instance.FreeCityTop:SetHide(false);
					end
                    
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------
                    -- My Mod UI 2
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------
                    
                    if warexhaustionlevel and warexhaustionlevel ~= 0 then                                                                     -- War Exhaustion
                        instance.WarExhaustionTop:SetHide(false);
                        SetPressureBreakdownColumn(instance.WarExhaustionPressureValue, instance.WarExhaustionFontIcon, Round(warexhaustionlevel,1));
                    else
                        instance.WarExhaustionTop:SetHide(true);
                    end
                    if ReligionMOdValue then                                                                                                  -- Religion
                    SetPressureBreakdownColumn(instance.ReligionPressureValue, instance.ReligionFontIcon, round(ReligionMOdValue,1));
                        local tooltip2:string = Locale.Lookup("Religion");
                        if tooltipReligionMod2 ~= "" then
                        tooltip2 = tooltip2 .. "[NEWLINE][NEWLINE]" .. Locale.Lookup(tooltipReligionMod2)
                            
                        end
                        instance.ReligionTop:SetToolTipString(tooltip2);
                    end
                    if GovernmentMod2 then                                                                                                      -- Government
                        SetPressureBreakdownColumn(instance.GovernmentModPressureValue, instance.GovernmentModFontIcon, round(GovernmentMod2,1));    
                        instance.GovernmentModTop:SetToolTipString(GovermentTooltip);
                    end
                    
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------
                    -- My Mod UI 2 end
                    ----------------------------------------------------------------------------------------------------------------------------------------------------------    
				end
				self.m_Instance.LoyaltyInfo.MainStack:CalculateSize();
				local newSizeY = self.m_Instance.LoyaltyInfo.MainStack:GetSizeY();
				self.m_Instance.LoyaltyInfo.CulturalIdentityExpandedButton:SetSizeY(newSizeY + 10);
				instance.Top:SetHide(false);
			end
		end
	end
end

function CityBanner:UpdateLoyaltyWarning()
	local pCity : table = self:GetCity();
	if (pCity == nil or self.m_Type ~= BANNERTYPE_CITY_CENTER) then
		return;
	end

	local pCulturalIdentity = pCity:GetCulturalIdentity();
	if (pCulturalIdentity ~= nil) then
		local eNextOwner = -1;
		if (self.m_IsCurrentlyVisible and not self.m_IsForceHide) then
			local nTurns:number = pCulturalIdentity:GetTurnsToConversion();
			local eOutcome:number = pCulturalIdentity:GetConversionOutcome();
			if (eOutcome == IdentityConversionOutcome.LOSING_LOYALTY and nTurns < 20) then
				eNextOwner = pCulturalIdentity:GetPotentialTransferPlayer();
			end
		end

		if (eNextOwner ~= self.m_eLoyaltyWarningPlayer) then
			local plot = Map.GetPlotIndex(self.m_PlotX, self.m_PlotY);

			-- clear the previous warning icon
			if (self.m_eLoyaltyWarningPlayer ~= -1) then
				UILens.ClearHex(LensLayers.LOYALTY_FREECITY_WARNING, plot);
			end
				
			-- create the new warning icon
			if (eNextOwner ~= -1) then
				local eLocalPlayer : number = Game.GetLocalPlayer(); 
				local kPlayerType = PlayerConfigurations[eNextOwner]:GetCivilizationTypeName();
				local nSecondaryColor, nPrimaryColor = UI.GetPlayerColors(eNextOwner);
				local nIconColor = (kPlayerType == "CIVILIZATION_FREE_CITIES") and nPrimaryColor or nSecondaryColor;
				local kAssetName = "LoyaltyWarning_" .. kPlayerType;
				-- TODO the lens model system should be expanded so this and the loyalty lens icons can use both civ colors
				UILens.SetLayerHexesColoredArea(LensLayers.LOYALTY_FREECITY_WARNING, eLocalPlayer, { plot }, nIconColor, kAssetName);
			end

			self.m_eLoyaltyWarningPlayer = eNextOwner;
		end
	end
end

-- ===========================================================================
function SetPressureBreakdownColumn(valueLabel:table, fontIconLabel:table, pressureValue:number)
	if pressureValue > 0 then
		valueLabel:SetText(Locale.Lookup("LOC_CULTURAL_IDENTITY_POSITIVE_PRESSURE", pressureValue));
		valueLabel:SetColorByName("White");
		fontIconLabel:SetText("[ICON_PressureUp]");
		fontIconLabel:SetHide(false);
	elseif pressureValue < 0 then
		valueLabel:SetText(pressureValue);
		valueLabel:SetColorByName("Red");
		fontIconLabel:SetText("[ICON_PressureDown]");
		fontIconLabel:SetHide(false);
	else
		valueLabel:SetText(Locale.Lookup("LOC_CULTURAL_IDENTITY_POSITIVE_PRESSURE", pressureValue));
		valueLabel:SetColorByName("Gray");
		fontIconLabel:SetHide(true);
	end
end

-- ===========================================================================
function CityBanner:UpdateSelected( state : boolean )
	local pCity : table = self:GetCity();
	if (pCity ~= nil) then
		UI.DeselectCity( pCity );
	end
end

-- ===========================================================================
function CityBanner:UpdatePosition()
	local yOffset = 0;	--offset for 2D strategic view
	local zOffset = 0;	--offset for 3D world view
	
	if (UI.GetWorldRenderView() == WorldRenderView.VIEW_2D) then
		yOffset = YOFFSET_2DVIEW;
		zOffset = 0;
	else
		yOffset = 0;
		yOffset = -25 + m_zoomMultiplier*25;
		zOffset = ZOFFSET_3DVIEW;
	end
	
	local worldX;
	local worldY;
	local worldZ;

	worldX, worldY, worldZ = UI.GridToWorld( self.m_PlotX, self.m_PlotY );
	self.m_Instance.Anchor:SetWorldPositionVal( worldX, worldY+yOffset, worldZ+zOffset );
end

-- ===========================================================================
function OnRefreshBannerPositions()
	--print("Refreshing banner positions");

	local pLocalPlayerVis = PlayersVisibility[Game.GetLocalPlayer()];
	if (pLocalPlayerVis ~= nil) then
		local players = Game.GetPlayers();
		for i, player in ipairs(players) do
			local playerID = player:GetID();
			local playerCities = players[i]:GetCities();
			for ii, city in playerCities:Members() do
				local cityID		:number = city:GetID();
				local locX			:number = city:GetX();
				local locY			:number = city:GetY();
				local isVisChange	:boolean = false;
				
				if pLocalPlayerVis:IsVisible(locX, locY) then
					OnCityVisibilityChanged(playerID, cityID, PLOT_VISIBLE);
					isVisChange = true;
				elseif pLocalPlayerVis:IsRevealed(locX, locY) then
					OnCityVisibilityChanged(playerID, cityID, PLOT_REVEALED);
					isVisChange = true;
				end

				local cityBanner:table = GetCityBanner( playerID, cityID );
				if (cityBanner ~= nil) then
					cityBanner:UpdatePosition();
				end
			end
			local playerDistricts = players[i]:GetDistricts();
			for ii, district in playerDistricts:Members() do
				local districtID = district:GetID();
				local locX = district:GetX();
				local locY = district:GetY();
				if (pLocalPlayerVis:IsVisible(locX, locY) == true) then
					OnDistrictVisibilityChanged(playerID, districtID, PLOT_VISIBLE);
					local miniBanner:table = GetMiniBanner( playerID, districtID );
					if (miniBanner ~= nil) then
						miniBanner:UpdatePosition();
					end
				end
			end
		end
	end
end

-- ===========================================================================
function CanRangeAttack(pCityOrDistrict : table)

	-- An invalid plot means we want to know if there are any locations that the city can range strike.

	return CityManager.CanStartCommand( pCityOrDistrict, CityCommandTypes.RANGE_ATTACK );
end

-- ===========================================================================
function CityBanner:UpdateRangeStrike()

	local controls:table = self.m_Instance;
	if controls.CityStrike == nil then
		-- This normal behaviour in the case of missile silo and aerodrome minibanners
		return;
	end

	local pDistrict:table = self:GetDistrict();
	if pDistrict ~= nil and self:IsTeam() then
		if (self.m_Player:GetID() == Game.GetLocalPlayer() and CanRangeAttack(pDistrict) ) then
			controls.CityStrike:SetHide(false);
		else
			controls.CityStrike:SetHide(true);
		end
	else
		-- are we looking at an Improvement miniBanner (Airstrip)?
		-- if so, just hide the attack container
		controls.CityStrike:SetHide(true);
	end
end

-- ===========================================================================
function OnCityStrikeButtonClick( playerID, cityID )
	local pPlayer = Players[playerID];
	if (pPlayer == nil) then
		return;
	end
	
	local pCity = pPlayer:GetCities():FindID(cityID);
	if (pCity == nil) then
		return;
	end;
	
	UI.SelectCity( pCity );
	UI.SetInterfaceMode(InterfaceModeTypes.CITY_RANGE_ATTACK);
end

-- ===========================================================================
function OnDistrictRangeStrikeButtonClick( playerID, districtID )
	local pPlayer = Players[playerID];
	if (pPlayer == nil) then
		return;
	end
	
	local pDistrict = pPlayer:GetDistricts():FindID(districtID);
	if (pDistrict == nil) then
		return;
	end;
	
	UI.DeselectAll();
	UI.SelectDistrict(pDistrict);
	UI.SetInterfaceMode(InterfaceModeTypes.DISTRICT_RANGE_ATTACK);
end

-- ===========================================================================
function OnICBMStrikeButtonClick( iPlotID, eWMD )
	local pPlot = Map.GetPlotByIndex(iPlotID);
	if (pPlot ~= nil) then
		local pCity = Cities.GetPlotPurchaseCity(pPlot);
		if (pCity ~= nil) then
			UI.SelectCity(pCity);
			local tParameters = {};
			tParameters[CityCommandTypes.PARAM_WMD_TYPE] = eWMD;
			tParameters[CityCommandTypes.PARAM_X0] = pPlot:GetX();
			tParameters[CityCommandTypes.PARAM_Y0] = pPlot:GetY();
			UI.SetInterfaceMode(InterfaceModeTypes.ICBM_STRIKE, tParameters);
		end
	end
end


-- ===========================================================================
function AddCityBannerToMap( playerID: number, cityID : number )
	local idLocalPlayer	:number = Game.GetLocalPlayer();
	local pPlayer		:table  = Players[playerID];

	local pCity = pPlayer:GetCities():FindID(cityID);
	if (pCity ~= nil) then
		local idDistrict = pCity:GetDistrictID();
		if (idLocalPlayer == playerID) then
			return CityBanner:new( playerID, cityID, idDistrict, BANNERTYPE_CITY_CENTER, BANNERSTYLE_LOCAL_TEAM );
		else
			return CityBanner:new( playerID, cityID, idDistrict, BANNERTYPE_CITY_CENTER, BANNERSTYLE_OTHER_TEAM );
		end
	end	
end

-- ===========================================================================
function DestroyCityBanner( playerID: number, cityID : number )
	local cityBanner:table = GetCityBanner( playerID, cityID );
	if (cityBanner ~= nil) then
		cityBanner:destroy();
		CityBannerInstances[ playerID ][ cityID ] = nil;
	end	
end

-- ===========================================================================
function AddMiniBannerToMap( playerID: number, cityID: number, districtID: number, styleEnum:number )
	local idLocalPlayer	:number = Game.GetLocalPlayer();
	local pPlayer		:table  = Players[playerID];

	if (idLocalPlayer == playerID) then		
		return CityBanner:new( playerID, cityID, districtID, styleEnum, BANNERSTYLE_LOCAL_TEAM );		
	else
		return CityBanner:new( playerID, cityID, districtID, styleEnum, BANNERSTYLE_OTHER_TEAM );		
	end
end

-- ===========================================================================
function OnCityAddedToMap( playerID: number, cityID : number, cityX : number, cityY : number )	
	if (CityBannerInstances[ playerID ] ~= nil and
	    CityBannerInstances[ playerID ][ cityID ] ~= nil) then
	    return;
    end
	AddCityBannerToMap( playerID, cityID );
end

-- ===========================================================================
function OnDistrictAddedToMap( playerID: number, districtID : number, cityID :number, districtX : number, districtY : number, districtType:number, percentComplete:number )

	local locX = districtX;
	local locY = districtY;
	local type = districtType;

	local pPlayer = Players[playerID];
	if (pPlayer ~= nil) then
		local pDistrict = pPlayer:GetDistricts():FindID(districtID);
		if (pDistrict ~= nil) then
			local pCity = pDistrict:GetCity();
			local cityID = pCity:GetID();
			-- It is possible that the city is not there yet. e.g. city-center district is placed, the city is placed immediately afterward.
			if (pCity ~= nil) then
				-- Is the district at the city? i.e. its a city-center?
				if (pCity:GetX() == pDistrict:GetX() and pCity:GetY() == pDistrict:GetY()) then
					-- Yes, just update the city banner with the district ID.
					local cityBanner:table = GetCityBanner( playerID, pCity:GetID() );
					if cityBanner then
						cityBanner.m_DistrictID = districtID;
						cityBanner:UpdateRangeStrike();
						cityBanner:UpdateStats();
						cityBanner:UpdateColor();
					end
				else
					-- Create a banner for a district that is not the city-center
					local miniBanner:table = GetMiniBanner( playerID, districtID );
					if (miniBanner == nil) then
						if (pDistrict:IsComplete()) then
							--only create district miniBanners for districts that are completed
							if ( GameInfo.Districts[pDistrict:GetType()].AirSlots > 0 ) then
								AddMiniBannerToMap( playerID, cityID, districtID, BANNERTYPE_AERODROME );
							elseif (pDistrict:GetDefenseStrength() > 0 ) then
								AddMiniBannerToMap( playerID, cityID, districtID, BANNERTYPE_ENCAMPMENT );
							end
						end
					else
						miniBanner:UpdateStats();
					end
				end
			end
		end
	end

end

-- ===========================================================================
function OnImprovementAddedToMap(locX, locY, eImprovementType, eOwner)

	if eImprovementType == -1 then
		UI.DataError("Received -1 eImprovementType for ("..tostring(locX)..","..tostring(locY)..") and owner "..tostring(eOwner));
		return;
	end

	local improvementData:table = GameInfo.Improvements[eImprovementType];

	if improvementData == nil then
		UI.DataError("No database entry for eImprovementType #"..tostring(eImprovementType).." for ("..tostring(locX)..","..tostring(locY)..") and owner "..tostring(eOwner));
		return;
	end
	
	-- Right now we're only interested in the Airstrip improvement
	if ( improvementData.AirSlots == 0 and improvementData.WeaponSlots == 0) then
		return;
	end

	local pPlayer:table = Players[eOwner];
	local localPlayerID:number = Game.GetLocalPlayer();
	if (pPlayer ~= nil) then
		local plotID = Map.GetPlotIndex(locX, locY);
		if (plotID ~= nil) then
			local miniBanner = GetMiniBanner( eOwner, plotID );
			if (miniBanner == nil) then
				if ( improvementData.AirSlots > 0 ) then
					--we're passing -1 as the cityID and the plotID as the districtID argument since Airstrips aren't associated with a city or a district
					AddMiniBannerToMap( eOwner, -1, plotID, BANNERTYPE_AERODROME );
				elseif ( improvementData.WeaponSlots > 0 ) then
					local ownerCity = Cities.GetPlotPurchaseCity(locX, locY);
					local cityID = ownerCity:GetID();
					-- we're passing the plotID as the districtID argument because we need the location of the improvement
					AddMiniBannerToMap( eOwner, cityID, plotID, BANNERTYPE_MISSILE_SILO );
				end
			else
				miniBanner:UpdateStats();
				miniBanner:UpdateColor();
			end
		end
	end
end

-- ===========================================================================
function OnDistrictProgressChanged(playerID: number, districtID : number, districtX : number, districtY : number, districtType:number, percentComplete:number)
	local pPlayer = Players[playerID];
	if (pPlayer ~= nil) then
		local pDistrict = pPlayer:GetDistricts():FindID(districtID);
		if (pDistrict ~= nil) then
			
		end
	end
end

-- ===========================================================================
function OnCityRemovedFromMap( playerID: number, cityID : number )
	DestroyCityBanner(playerID, cityID);
end

-- ===========================================================================
function OnDistrictRemovedFromMap( playerID : number, districtID : number )
	local bannerInstance = GetMiniBanner(playerID, districtID);
	if (bannerInstance ~= nil) then
		bannerInstance:destroy();
		MiniBannerInstances[playerID][districtID] = nil;
	end
end

-- ===========================================================================
function OnImprovementRemovedFromMap( locX :number, locY :number, eOwner :number )
	local plotID = Map.GetPlotIndex(locX, locY);
	if (plotID > 0) then
		local bannerInstance = GetMiniBanner( eOwner, plotID );
		if (bannerInstance ~= nil) then
			bannerInstance:destroy();
			bannerInstance:SetFogState( eVisibility );
		end
	end
end

-- ===========================================================================
function OnCityVisibilityChanged( playerID: number, cityID : number, eVisibility : number)
	local bannerInstance:table = GetCityBanner( playerID, cityID );
	if bannerInstance then
		bannerInstance:SetFogState( eVisibility );
	end
end

-- ===========================================================================
function OnCityOccupationChanged( playerID: number, cityID : number )
	RefreshBanner( playerID, cityID );
end

-- ===========================================================================
function OnCityPopulationChanged( playerID: number, cityID : number )
	RefreshBanner( playerID, cityID );
end

-- ===========================================================================
function OnDistrictVisibilityChanged( playerID :number, districtID :number, eVisibility :number )
	local bannerInstance = GetMiniBanner( playerID, districtID );
	if (bannerInstance ~= nil) then
		bannerInstance:SetFogState( eVisibility );
	end
end

-- ===========================================================================
function OnImprovementVisibilityChanged( locX :number, locY :number, eImprovementType :number, eVisibility :number )
	if ( eImprovementType == -1 ) then
		return;
	end
	-- We're only interested in the Airstrip or Missile Silo improvements
	if ( GameInfo.Improvements[eImprovementType].AirSlots > 0 or GameInfo.Improvements[eImprovementType].WeaponSlots > 0) then
		local plotID = Map.GetPlotIndex(locX, locY);
		if (plotID > 0) then
			local plot = Map.GetPlotByIndex(plotID);
			if (plot ~= nil) then
				local x = plot:GetX();
				local y = plot:GetY();
				local playerID = plot:GetImprovementOwner();
				local bannerInstance = GetMiniBanner( playerID, plotID );
				if (bannerInstance ~= nil) then
					bannerInstance:SetFogState( eVisibility );
				end
			end
		end
	else
		return;
	end
end

-- ===========================================================================
function OnBuildingChanged( plotX:number, plotY:number, buildingIndex:number, playerID:number, iPercentComplete:number)
	
	local pPlayer = Players[playerID];
	if (pPlayer ~= nil and pPlayer:GetCities() ~= nil) then
		
		-- Update the capital, since for now capital status is shown in name
		local pCapital = pPlayer:GetCities():GetCapitalCity();
		if (pCapital ~= nil) then
			local cityBanner:table = GetCityBanner( playerID, pCapital:GetID() );
			if cityBanner then
				cityBanner:UpdateName();
			end
		end

		-- Update the city defenses UI if walls were constructed
		if (playerID == Game.GetLocalPlayer()) then
			local pCity = CityManager.GetCityAt(plotX, plotY);
			if (pCity ~= nil) then
				local cityBanner:table = GetCityBanner( playerID, pCity:GetID() );
				if cityBanner then
					cityBanner:UpdateRangeStrike();
				end
			end
		end
	end

end

-- ===========================================================================
function OnCityNameChange( playerID: number, cityID : number)
	
	local banner:table = GetCityBanner( playerID, cityID );
	if (banner ~= nil ) then
		banner:UpdateName();   
    end

end

-- ===========================================================================
function OnCapitalCityChanged( playerID: number, cityID : number )
	-- Ensure not in autoplay
	if Game.GetLocalPlayer() < 0 then
		return;
	end

    local banner:table = GetCityBanner( playerID, cityID );
	if (banner ~= nil ) then
		banner:UpdateName();   
    end
end

-- ===========================================================================
function OnCityReligionChanged( playerID: number, cityID : number, eVisibility : number, city)

	-- Ensure not in autoplay
	if Game.GetLocalPlayer() < 0 then
		return;
	end

    local banner:table = GetCityBanner( playerID, cityID );
	if (banner ~= nil and banner.m_Instance.ReligionInfo ~= nil and banner:IsVisible()) then
		banner:UpdateReligion();   -- For now religion is shown in name
    end
end

-- ===========================================================================
function OnQuestChanged( fromPlayerID:number, toPlayerID:number)

	-- Update the capital of the player the quest is from
	local pFromPlayer = Players[fromPlayerID];
	if (pFromPlayer ~= nil and pFromPlayer:GetCities() ~= nil) then
		local pCapital = pFromPlayer:GetCities():GetCapitalCity();
		if (pCapital ~= nil) then
			local bannerInstance = GetCityBanner( fromPlayerID, pCapital:GetID() );
			if (bannerInstance ~= nil) then
				bannerInstance:UpdateName();
			end
		end
	end
end

-- ===========================================================================
function OnDistrictCombatChanged(eventSubType, playerID, districtID)
	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then
		local pDistrict = pPlayer:GetDistricts():FindID(districtID);
		if (pDistrict ~= nil) then
			local pCity = pDistrict:GetCity();
			local banner = GetCityBanner(playerID, pCity:GetID());
			if (banner ~= nil) then
				banner:UpdateRangeStrike();
				banner:UpdateStats();
			end

			local miniBanner = GetMiniBanner(playerID, districtID);
			if (miniBanner ~= nil) then
				miniBanner:UpdateRangeStrike();
				miniBanner:UpdateStats();
			end
		end
    end
end

-- ===========================================================================
function OnCityDefenseStatusChanged(playerID, iValue)
	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then
		local pPlayerDistricts:table = pPlayer:GetDistricts();
		for _, district in pPlayerDistricts:Members() do
			local pCity = district:GetCity();
			local districtID = district:GetID();
			if (district:GetX() == pCity:GetX() and district:GetY() == pCity:GetY()) then
				local banner = GetCityBanner(playerID, pCity:GetID());
				if (banner ~= nil) then
					banner:UpdateRangeStrike();
					banner:UpdateStats();
				end
			else
				local miniBanner = GetMiniBanner(playerID, districtID);
				if (miniBanner ~= nil) then
					miniBanner:UpdateRangeStrike();
					miniBanner:UpdateStats();
				end
			end
		end
	end
end

-- ===========================================================================
function OnDistrictDamageChanged( playerID:number, districtID:number, damageType:number, newDamage:number, oldDamage:number)
	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then
		local pDistrict = pPlayer:GetDistricts():FindID(districtID);
		if (pDistrict ~= nil) then
			local pCity = pDistrict:GetCity();
			if (pDistrict:GetX() == pCity:GetX() and pDistrict:GetY() == pCity:GetY()) then
				local banner = GetCityBanner(playerID, pCity:GetID());
				if (banner ~= nil) then
					banner:UpdateStats();
				end
			else
				local miniBanner = GetMiniBanner(playerID, districtID);
				if (miniBanner ~= nil) then
					miniBanner:UpdateStats();
				end
			end

			-- Add the world space text to show the delta for the damage.
			-- Can the local team see the plot where the district is?
			local pLocalPlayerVis = PlayersVisibility[Game.GetLocalPlayer()];
			if (pLocalPlayerVis ~= nil) then
				if (pLocalPlayerVis:IsVisible(pDistrict:GetX(), pDistrict:GetY())) then
								
					local iDelta = newDamage - oldDamage;
					local szText;

					if (damageType == DefenseTypes.DISTRICT_GARRISON) then
						if (iDelta < 0) then
							szText = Locale.Lookup("LOC_WORLD_DISTRICT_GARRISON_DAMAGE_DECREASE_FLOATER", -iDelta);
						else
							szText = Locale.Lookup("LOC_WORLD_DISTRICT_GARRISON_DAMAGE_INCREASE_FLOATER", -iDelta);
						end
					elseif (damageType == DefenseTypes.DISTRICT_OUTER) then
						if (iDelta < 0) then
							szText = Locale.Lookup("LOC_WORLD_DISTRICT_DEFENSE_DAMAGE_DECREASE_FLOATER", -iDelta);
						else
							szText = Locale.Lookup("LOC_WORLD_DISTRICT_DEFENSE_DAMAGE_INCREASE_FLOATER", -iDelta);
						end
					end


					UI.AddWorldViewText(EventSubTypes.DAMAGE, szText, pDistrict:GetX(), pDistrict:GetY(), 0);
				end
			end
		end
	end
	-- print("A District has been damaged");
	-- print(playerID, districtID, outerDamage, garrisonDamage);
end


-- ===========================================================================
function UpdateStats( playerID:number, cityID:number )
	if (playerID == Game.GetLocalPlayer()) then
		local pPlayer = Players[ playerID ];
		if (pPlayer ~= nil) then
			local pCity = pPlayer:GetCities():FindID(cityID);
			if (pCity ~= nil) then
				local banner = GetCityBanner(playerID, cityID);
				if (banner ~= nil) then
					banner:UpdateStats();
				end
			end
			-- Update minibanners associated with the given city
			local playerMiniBannerInstances = MiniBannerInstances[ playerID ];
			if (playerMiniBannerInstances ~= nil) then
				for id, banner in pairs(playerMiniBannerInstances) do
					if (banner ~= nil and banner.m_CityID == cityID) then
						banner:UpdateStats();
					end
				end
			end
		end
	end
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnCityFocusChange( playerID:number, cityID:number )
	UpdateStats( playerID, cityID );
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnCityProductionChanged( playerID:number, cityID:number)
	UpdateStats( playerID, cityID );
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnCityProductionUpdate( playerID:number, cityID:number)
	UpdateStats( playerID, cityID );
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnCityProductionCompleted( playerID:number, cityID:number)
	UpdateStats( playerID, cityID );
end


-- ===========================================================================
--	Update stats and button to attack on banners
-- ===========================================================================
function RefreshPlayerBanners( playerID:number )
	if playerID == -1 then return; end

	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then

		if (CityBannerInstances[ playerID ] == nil) then
			return;
		end
		local playerCityBannerInstances = CityBannerInstances[ playerID ];
		for id, banner in pairs(playerCityBannerInstances) do
			if (banner ~= nil) then
				banner:UpdateStats();
				banner:UpdateRangeStrike();
			end
		end

		if (MiniBannerInstances[ playerID ] == nil) then
			return;
		end
		local playerMiniBannerInstances = MiniBannerInstances[ playerID ];
		for id, banner in pairs(playerMiniBannerInstances) do
			if (banner ~= nil) then
				banner:UpdateStats();
				banner:UpdateRangeStrike();
			end
		end
	end

end

-- ===========================================================================
function RefreshPlayerRangeStrike( playerID:number )
	if playerID == -1 then return; end

	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then

		if (CityBannerInstances[ playerID ] == nil) then
			return;
		end
		local playerCityBannerInstances = CityBannerInstances[ playerID ];
		for id, banner in pairs(playerCityBannerInstances) do
			if (banner ~= nil) then
				banner:UpdateRangeStrike();
			end
		end

		if (MiniBannerInstances[ playerID ] == nil) then
			return;
		end
		local playerMiniBannerInstances = MiniBannerInstances[ playerID ];
		for id, banner in pairs(playerMiniBannerInstances) do
			if (banner ~= nil) then
				banner:UpdateRangeStrike();
			end
		end
	end

end

-- ===========================================================================
function RefreshBanner( playerID:number, cityID:number )
	local banner = GetCityBanner(playerID, cityID);
	if (banner ~= nil) then
		banner:UpdateStats();
		banner:UpdateRangeStrike();
		banner:UpdateName();
	end
end

-- ===========================================================================
function RefreshMiniBanner( playerID:number, districtID:number )
	local banner = GetMiniBanner(playerID, districtID);
	if (banner ~= nil) then
		banner:UpdateStats();
		banner:UpdateRangeStrike();
	end
end

-- ===========================================================================
function OnCityUnitsChanged( playerID:number, cityID:number )
	if playerID == Game.GetLocalPlayer() then
		RefreshBanner( playerID, cityID );
	end
end

-- ===========================================================================
function OnDistrictUnitsChanged( playerID:number, districtID:number )
	if playerID == Game.GetLocalPlayer() then
		RefreshMiniBanner( playerID, districtID );
	end
end

-- ===========================================================================
function OnSiegeStatusChanged( playerID:number, cityID:number, bIsBesieged:boolean )
	if (playerID == -1) then
		return;
	end

	RefreshBanner( playerID, cityID );
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnUnitMoved( playerID:number, unitID:number )
	local localPlayer = Game.GetLocalPlayer();
	if localPlayer ~= -1 and localPlayer ~= playerID and Players[localPlayer]:IsTurnActive() then
		m_refreshLocalPlayerRangeStrike = true;
	end
end

function FlushChanges()
	if m_refreshLocalPlayerRangeStrike then
		RefreshPlayerRangeStrike( Game.GetLocalPlayer() );
		m_refreshLocalPlayerRangeStrike = false;
	end
end

-- ===========================================================================
function OnUnitAddedOrUpgraded( playerID:number, unitID:number )
	-- Update city and district garrison strength values if a melee unit has been added or upgraded.
	-- This is done because the base city strength is calculated using the max melee strength for the player.
	local localPlayer = Game.GetLocalPlayer();
	if localPlayer == -1 or Players[localPlayer]:IsTurnActive() then -- Don't do this during end turn times
		local pUnit = Players[ playerID ]:GetUnits():FindID(unitID);
		if pUnit ~= nil then
			local pUnitDef = GameInfo.Units[pUnit:GetUnitType()];
			if pUnitDef ~= nil then
				if pUnitDef.Combat > 0 then -- Only do this for melee units
					RefreshPlayerBanners( playerID );
				end
			end
		end
	end
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnUnitAddedToMap( playerID:number, unitID:number )
	OnUnitMoved( playerID, unitID );
	OnUnitAddedOrUpgraded( playerID, unitID );
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnUnitRemovedFromMap( playerID:number, unitID:number )
	OnUnitMoved( playerID, unitID );
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnUnitUpgraded( playerID:number, unitID:number )
	OnUnitAddedOrUpgraded( playerID, unitID );
end

-- ===========================================================================
--	Game Event
-- ===========================================================================
function OnDiplomacyDeclareWar( firstPlayerID:number, secondPlayerID:number )
	local localPlayer = Game.GetLocalPlayer();
	if firstPlayerID == localPlayer or secondPlayerID == localPlayer then
		m_refreshLocalPlayerRangeStrike = true;
	end
end

-- ===========================================================================
--	Game Event
-- ===========================================================================
function OnDiplomacyMakePeace( firstPlayerID:number, secondPlayerID:number )
	local localPlayer = Game.GetLocalPlayer();
	if firstPlayerID == localPlayer or secondPlayerID == localPlayer then
		m_refreshLocalPlayerRangeStrike = true;
	end
end

-- ===========================================================================
function OnWMDCountChanged( playerID:number, eWMD:number )
	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then
		if (MiniBannerInstances[ playerID ] == nil) then
			return;
		end
		local playerMiniBannerInstances = MiniBannerInstances[ playerID ];
		for id, banner in pairs(playerMiniBannerInstances) do
			if (banner ~= nil) then
				banner:UpdateStats();
			end
		end
	end
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnTurnActivated( playerID:number )
	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil) then
	
		local playerBannerInstances = CityBannerInstances[ playerID ];
		if (playerBannerInstances ~= nil) then
			for id, banner in pairs(playerBannerInstances) do
				if (banner ~= nil) then
					banner:UpdateStats();
				end
			end
		end

		local playerMiniBannerInstances = MiniBannerInstances[ playerID ];
		if (playerMiniBannerInstances ~= nil) then
			for id, banner in pairs(playerMiniBannerInstances) do
				if (banner ~= nil) then
					banner:UpdateStats();
				end
			end
		end
	end
end

-- ===========================================================================
--	Game Engine Event
-- ===========================================================================
function OnPolicyChanged( playerID:number )
	RefreshPlayerBanners( playerID );
end

function OnPlayerAgeChanged ( playerID:number )
	local pPlayer = Players[ playerID ];
	if (pPlayer ~= nil and pPlayer:IsHuman() == true) then
		RefreshPlayerBanners( playerID );
	end
end

-- ===========================================================================
function OnSpyMissionCompleted( playerID:number, missionID:number )
	-- When a spy mission completes update governors of cities with spies in case they've been neutralized
	local pPlayer:table = Players[playerID];
	if pPlayer then
		local pPlayerUnits:table = pPlayer:GetUnits();
		for i, unit in pPlayerUnits:Members() do
			local unitInfo:table = GameInfo.Units[unit:GetUnitType()];
			if unitInfo.Spy then
				local pCity = CityManager.GetCityAt(unit:GetX(), unit:GetY());
				if pCity ~= nil then
					local cityBanner:table = GetCityBanner(pCity:GetOwner(), pCity:GetID());
					if cityBanner then
						cityBanner:UpdateGovernor(pCity);
					end
				end
			end
		end
	end
end

-- ===========================================================================
--	Reload all the content
-- ===========================================================================
function Reload()

	local pLocalPlayerVis:table = PlayersVisibility[Game.GetLocalPlayer()];
	if pLocalPlayerVis ~= nil then
		local players = Game.GetPlayers();
		for i, player in ipairs(players) do
			local playerID		:number = player:GetID();
			local pPlayerCities	:table = players[i]:GetCities();

			for _, city in pPlayerCities:Members() do
				local cityID:number = city:GetID();
				local locX	:number = city:GetX();
				local locY	:number = city:GetY();
				OnCityAddedToMap( playerID, cityID, locX, locY );
				if (pLocalPlayerVis:IsVisible(locX, locY) == true) then
					OnCityVisibilityChanged(playerID, cityID, PLOT_VISIBLE);
				end

			end

			local pPlayerDistricts:table = players[i]:GetDistricts();
			for _, district in pPlayerDistricts:Members() do
				local districtID = district:GetID();
				local locX = district:GetX();
				local locY = district:GetY();
				OnDistrictAddedToMap( playerID, districtID, locX, locY );
				if (pLocalPlayerVis:IsVisible(locX, locY) == true) then
					OnDistrictVisibilityChanged(playerID, districtID, PLOT_VISIBLE);
				end
			end

			local pPlayerImprovements = players[i]:GetImprovements();
			if (pPlayerImprovements ~= nil) then
				local tImprovementLocations:table = pPlayerImprovements:GetImprovementPlots();
				for _, plotID in ipairs(tImprovementLocations) do
					local pPlot = Map.GetPlotByIndex(plotID);
					if (pPlot ~= nil) then
						local eImprovement = pPlot:GetImprovementType();
						if (eImprovement >= 0) then
							local locX = pPlot:GetX();
							local locY = pPlot:GetY();
							OnImprovementAddedToMap(locX, locY, eImprovement, playerID);
							if (pLocalPlayerVis:IsVisible(locX, locY) == true) then
								OnImprovementVisibilityChanged(locX, locY, eImprovement, PLOT_VISIBLE); 
							end
						end
					end
				end
			end

		end
	end
end

-- ===========================================================================
function OnEventPlaybackComplete()

	for playerID, cityID in m_pDirtyCityComponents:Members() do
		local banner = GetCityBanner(playerID, cityID);
		if (banner ~= nil) then
			banner:UpdateStats();
		end
	end

	m_pDirtyCityComponents:Clear();
end

----------------------------------------------------------------
function OnLocalPlayerChanged( localPlayerID:number , prevLocalPlayerID:number )

	-- Hide all the flags, we will get updates later
	for _, playerBannerInstances in pairs(CityBannerInstances) do
		for id, banner in pairs(playerBannerInstances) do
			if (banner ~= nil) then
				banner:SetHide(true);
			end
		end
    end

	for _, playerBannerInstances in pairs(MiniBannerInstances) do
		for id, banner in pairs(playerBannerInstances) do
			if (banner ~= nil) then
				banner:SetHide(true);
			end
		end
	end

	--	Rebuild all city banner instances in the context of the new local player.
	for iPlayer,kCityBanners in pairs(CityBannerInstances) do
		for iCity,kCityBanner in pairs(kCityBanners) do
			DestroyCityBanner( iPlayer, iCity );
			AddCityBannerToMap( iPlayer, iCity );
		end
	end

	for iPlayer,kMiniBanners in pairs(MiniBannerInstances) do
		for iMini,kMiniBanner in pairs(kMiniBanners) do
			local districtID:number = kMiniBanner.m_DistrictID;
			local typeID	:number = kMiniBanner.m_Type;
			local cityID	:number = kMiniBanner.m_CityID;
			kMiniBanner:destroy();
			AddMiniBannerToMap( iPlayer, cityID, districtID, typeID );
		end
	end
end

-- ===========================================================================
--	Game Engine EVENT
-- ===========================================================================
function OnPlayerTurnActivated(player, isFirstTimeThisTurn)
	-- PlayerTurnActivated is post DoTurn processing for the beginning of the turn.
	if (isFirstTimeThisTurn and Game.GetLocalPlayer() == player) then
		OnRefreshBannerPositions();						-- Ensure visibility is correctly set.
		RefreshPlayerBanners( player );
	end
end

-- ===========================================================================
function OnObjectPairingChanged(eSubType, parentOwner, parentType, parentID, childOwner, childType, childID)
	local pPlayer = Players[ parentOwner ];
	if (pPlayer ~= nil) then

		local bannerInstance = GetCityBanner( parentOwner, parentID );
		if (bannerInstance ~= nil) then
			bannerInstance:UpdateStats( bannerInstance );
		end

		local miniBannerInstance = GetMiniBanner( parentOwner, parentID );
		if (miniBannerInstance ~= nil) then
			miniBannerInstance:UpdateStats( miniBannerInstance );
		end

	end
end

-- ===========================================================================
function RegisterDirtyEvents()	
	m_pDirtyCityComponents = DirtyComponentsManager.Create();
	m_pDirtyCityComponents:AddEvent("CITY_POPULATION_CHANGED");
	m_pDirtyCityComponents:AddEvent("CITY_RELIGION_CHANGED");
end

-- ===========================================================================
function RealizeReligion()
	
	m_HolySiteIconsIM:ResetInstances();
	-- Only clear the religion lens if we're turning off lenses altogether, but not if switching to another modal lens. (Turning on another modal lens clears it already)
	if UI.GetInterfaceMode() ~= InterfaceModeTypes.VIEW_MODAL_LENS then
		UILens.ClearLayerHexes( LensLayers.HEX_COLORING_RELIGION );
	end
	
	for _, playerBannerInstances in pairs(CityBannerInstances) do
		for id, banner in pairs(playerBannerInstances) do
			if (banner ~= nil and banner.m_Instance.ReligionInfo ~= nil and banner:IsVisible()) then
				banner:UpdateReligion();
				banner:UpdatePosition();
			end
		end
	end
end

-- ===========================================================================
function RealizeLoyalty()
	for _, playerBannerInstances in pairs(CityBannerInstances) do
		for id, banner in pairs(playerBannerInstances) do
			if (banner ~= nil and banner.m_Instance.LoyaltyInfo and banner:IsVisible()) then
				banner:UpdateInfo(banner:GetCity());
				banner:UpdateLoyalty();
				banner:UpdatePosition();
			end
		end
	end
end

-- ===========================================================================
function OnContextInitialize( isHotload : boolean )
	if isHotload then
		LuaEvents.GameDebug_GetValues( "CityBannerManager" );
		Reload();
	end
end

-- ===========================================================================
--	Handle the UI shutting down.
function OnShutdown()
	-- Cache value for hotloading...
	LuaEvents.GameDebug_AddValue("CityBannerManager", "m_isLoyaltyLensActive", m_isLoyaltyLensActive);
	LuaEvents.GameDebug_AddValue("CityBannerManager", "m_isReligionLensActive", m_isReligionLensActive);
	
	DirtyComponentsManager.Destroy( m_pDirtyCityComponents );
	m_pDirtyCityComponents = nil;
end

-- ===========================================================================
--	LUA Event
--	Set cached values back after a hotload.
-- ===========================================================================
function OnGameDebugReturn( context:string, contextTable:table )
	if context == "CityBannerManager" then
		m_isLoyaltyLensActive = contextTable["m_isLoyaltyLensActive"];
		m_isReligionLensActive = contextTable["m_isReligionLensActive"];
		RealizeReligion();
		RealizeLoyalty();
	end
end

-- ===========================================================================
function OnBeginWonderReveal()
	ContextPtr:SetHide( true );
end

-- ===========================================================================
function OnEndWonderReveal()
	ContextPtr:SetHide( false );
end

-- ===========================================================================
--	Gamecore Event
--	Called once per layer that is turned on when a new lens is activated,
--	or when a player explicitly turns off the layer from the "player" lens.
-- ===========================================================================
function OnLensLayerOn( layerNum:number )
	if layerNum == LensLayers.HEX_COLORING_RELIGION then
		m_isReligionLensActive = true;
		RealizeReligion();
	elseif layerNum == LensLayers.CULTURAL_IDENTITY_LENS then
		m_isLoyaltyLensActive = true;
		RealizeLoyalty();
	end
end

-- ===========================================================================
--	Gamecore Event
--	Called once per layer that is turned on when a new lens is deactivated,
--	or when a player explicitly turns off the layer from the "player" lens.
-- ===========================================================================
function OnLensLayerOff( layerNum:number )
	if	layerNum == LensLayers.HEX_COLORING_RELIGION then
		m_isReligionLensActive = false;
		RealizeReligion();
	elseif layerNum == LensLayers.CULTURAL_IDENTITY_LENS then
		m_isLoyaltyLensActive = false;
		RealizeLoyalty();
	end
end

-- ===========================================================================
function OnSelectionChanged(owner, ID, i, j, k, bSelected, bEditable)
	banner = GetCityBanner(owner, ID);
	-- OnSelectionChanged event can only change one banner at a time
	if banner and owner == Game.GetLocalPlayer() then
		banner.m_IsSelected = bSelected;
		banner:UpdateColor();
	end
end

function OnCameraUpdate( vFocusX:number, vFocusY:number, fZoomLevel:number )

	-- If no change in the zoom, no update necessary.
	if (1-fZoomLevel) == m_zoomMultiplier then
		return;
	end
	m_zoomMultiplier= 1-fZoomLevel;
	
	if ( math.abs(m_zoomMultiplier-m_prevZoomMultiplier) > ZOOM_MULT_DELTA) then
		OnRefreshBannerPositions();
	end
	m_prevZoomMultiplier = m_zoomMultiplier;
end

-- ===========================================================================
function OnInterfaceModeChanged( oldMode:number, newMode:number )
	if newMode == InterfaceModeTypes.MAKE_TRADE_ROUTE or oldMode == InterfaceModeTypes.MAKE_TRADE_ROUTE then

		m_isTradeSelectionActive = newMode == InterfaceModeTypes.MAKE_TRADE_ROUTE;

		-- Show trading post icons on cities that contain a trading post with the local player
		for _, playerBannerInstances in pairs(CityBannerInstances) do
			for id, banner in pairs(playerBannerInstances) do
				if banner ~= nil then
					banner:UpdateDetails();
				end
			end
		end
	end
end

-- ===========================================================================
function OnCityLoyaltyChanged( playerID: number, cityID: number )
	local cityBanner:table = GetCityBanner(playerID, cityID);
	if cityBanner then
		cityBanner:UpdateInfo(cityBanner:GetCity());
		cityBanner:UpdateLoyalty();
	end
end

-- ===========================================================================
function OnCulturalIdentityConversionOutcomeChanged( playerID: number, cityID: number, conversionOutcome:number )
	local cityBanner:table = GetCityBanner(playerID, cityID);
	if cityBanner then
		cityBanner:UpdateInfo(cityBanner:GetCity());
		cityBanner:UpdateLoyalty();
	end
end

-- ===========================================================================
function OnGovernorChanged( playerID: number, governorID: number )
	local pPlayer = Players[playerID];
	local pGovernors = pPlayer:GetGovernors();
	local pGovernor = pGovernors:GetGovernor(GameInfo.Governors[governorID].Hash);

	if pGovernor then
		local pCity:table = pGovernor:GetAssignedCity();
		if pCity then
			local cityBanner:table = GetCityBanner(pCity:GetOwner(), pCity:GetID());
			if cityBanner then
				cityBanner:UpdateStats();
				cityBanner:UpdateLoyalty();
			end
		end
	end
end

-- ===========================================================================
function OnGovernorAssigned( playerID: number, governorID: number, cityOwner: number, cityID: number )
	local cityBanner:table = GetCityBanner(cityOwner, cityID);
	for _, playerBannerInstances in pairs(CityBannerInstances) do
		for id, banner in pairs(playerBannerInstances) do
			if (banner ~= nil and banner:IsVisible()) then
				banner:UpdateStats();
				banner:UpdateLoyalty();
			end
		end
	end
end

function Initialize()

	RegisterDirtyEvents();

	ContextPtr:SetInitHandler( OnContextInitialize );
	ContextPtr:SetShutdown( OnShutdown );

	Events.BeginWonderReveal.Add(				OnBeginWonderReveal );
	Events.BuildingChanged.Add(					OnBuildingChanged);
	Events.Camera_Updated.Add(					OnCameraUpdate );
	Events.CapitalCityChanged.Add(				OnCapitalCityChanged);
	Events.CityAddedToMap.Add(					OnCityAddedToMap );
	Events.CityDefenseStatusChanged.Add(		OnCityDefenseStatusChanged );
	Events.CityFocusChanged.Add(				OnCityFocusChange );
	Events.CityNameChanged.Add(					OnCityNameChange );
	Events.CityProductionChanged.Add(			OnCityProductionChanged);
	Events.CityProductionUpdated.Add(			OnCityProductionUpdate); 
	Events.CityProductionCompleted.Add(			OnCityProductionCompleted);
	Events.CityReligionChanged.Add(				OnCityReligionChanged );
	Events.CityReligionFollowersChanged.Add(	OnCityReligionChanged );
	Events.CityRemovedFromMap.Add(				OnCityRemovedFromMap );
	Events.CitySelectionChanged.Add(			OnSelectionChanged );
	Events.CityUnitsChanged.Add(                OnCityUnitsChanged );
	Events.CityVisibilityChanged.Add(			OnCityVisibilityChanged );
	Events.CityOccupationChanged.Add(			OnCityOccupationChanged );
	Events.CityPopulationChanged.Add(			OnCityPopulationChanged );
	Events.DiplomacyDeclareWar.Add(				OnDiplomacyDeclareWar );
	Events.DiplomacyMakePeace.Add(				OnDiplomacyMakePeace );
	Events.DistrictAddedToMap.Add(				OnDistrictAddedToMap );
	Events.DistrictBuildProgressChanged.Add(	OnDistrictAddedToMap);
	--Events.DistrictBuildProgressChanged.Add(	OnDistrictProgressChanged);
	Events.DistrictCombatChanged.Add(			OnDistrictCombatChanged );
	Events.DistrictDamageChanged.Add(			OnDistrictDamageChanged );	
	Events.DistrictRemovedFromMap.Add(			OnDistrictRemovedFromMap );
	Events.DistrictUnitsChanged.Add(			OnDistrictUnitsChanged );
	Events.DistrictVisibilityChanged.Add(		OnDistrictVisibilityChanged );
	Events.EndWonderReveal.Add(					OnEndWonderReveal );
	Events.GameCoreEventPlaybackComplete.Add(	OnEventPlaybackComplete);
	Events.ImprovementAddedToMap.Add(			OnImprovementAddedToMap );
	Events.ImprovementRemovedFromMap.Add(		OnImprovementRemovedFromMap );
	Events.ImprovementVisibilityChanged.Add(	OnImprovementVisibilityChanged );
	Events.InterfaceModeChanged.Add(			OnInterfaceModeChanged );
	Events.LensLayerOff.Add(					OnLensLayerOff );
	Events.LensLayerOn.Add(						OnLensLayerOn );
	Events.LocalPlayerChanged.Add(				OnLocalPlayerChanged);	
	Events.PlayerTurnActivated.Add(				OnPlayerTurnActivated);
	Events.ObjectPairing.Add(					OnObjectPairingChanged);
	Events.QuestChanged.Add(					OnQuestChanged );
	Events.UnitAddedToMap.Add(					OnUnitAddedToMap );
	Events.UnitMoved.Add(						OnUnitMoved );
	Events.UnitRemovedFromMap.Add(				OnUnitRemovedFromMap );
	Events.UnitUpgraded.Add(					OnUnitUpgraded );
	Events.UnitVisibilityChanged.Add(			OnUnitMoved );
	Events.WorldRenderViewChanged.Add(			OnRefreshBannerPositions);
	Events.WMDCountChanged.Add(					OnWMDCountChanged);
	Events.PlayerTurnActivated.Add(             OnTurnActivated);
	Events.GovernmentPolicyChanged.Add(         OnPolicyChanged );
	Events.GovernmentPolicyObsoleted.Add(       OnPolicyChanged );
	Events.SpyMissionCompleted.Add(				OnSpyMissionCompleted );
	Events.PlayerAgeChanged.Add(         OnPlayerAgeChanged );
	Events.PlayerDarkAgeChanged.Add(       OnPlayerAgeChanged );
	Events.CitySiegeStatusChanged.Add(			OnSiegeStatusChanged);
	Events.GameCoreEventPublishComplete.Add(	FlushChanges); --This event is raised directly after a series of gamecore events.

	-- Expansion1 related events
	Events.GovernorChanged.Add(					OnGovernorChanged);
	Events.GovernorAssigned.Add(				OnGovernorAssigned);
	Events.GovernorPromoted.Add(				OnGovernorChanged);
	Events.CityLoyaltyChanged.Add(				OnCityLoyaltyChanged);
	Events.CulturalIdentityConversionOutcomeChanged.Add( OnCulturalIdentityConversionOutcomeChanged);

	LuaEvents.GameDebug_Return.Add(OnGameDebugReturn);
end
Initialize();

